"""gNUMA differentiation kernel, wrapped up as a test."""

__copyright__ = "Copyright (C) 2015 Andreas Kloeckner, Lucas Wilcox"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import pytest

import pyopencl as cl  # noqa: F401
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import loopy as lp
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)


@pytest.mark.parametrize("ilp_multiple", [1, 2])
@pytest.mark.parametrize("Nq", [7])
@pytest.mark.parametrize("opt_level", [11])
def test_gnuma_horiz_kernel(ctx_factory, ilp_multiple, Nq, opt_level):  # noqa: N803
    pytest.importorskip("fparser")
    ctx = ctx_factory()

    if (ctx.devices[0].platform.name == "Portable Computing Language"
            and ilp_multiple > 1):
        # about 400s, cf. https://gitlab.tiker.net/inducer/loopy/-/jobs/421250#L937
        pytest.skip("takes a very long time to compile on pocl")

    import os

    filename = os.path.join(os.path.dirname(__file__), "strongVolumeKernels.f90")
    with open(filename) as sourcef:
        source = sourcef.read()

    source = source.replace("datafloat", "real*4")

    program = lp.parse_fortran(source, filename, seq_dependencies=False)

    hsv_r, hsv_s = program["strongVolumeKernelR"], program["strongVolumeKernelS"]

    hsv_r = lp.tag_instructions(hsv_r, "rknl")
    hsv_s = lp.tag_instructions(hsv_s, "sknl")
    hsv = lp.fuse_kernels([hsv_r, hsv_s], ["_r", "_s"])
    # hsv = hsv_s
    hsv = lp.add_nosync(hsv, "any", "writes:rhsQ", "writes:rhsQ", force=True)

    from gnuma_loopy_transforms import (
        fix_euler_parameters,
        set_D_storage_format,
        set_q_storage_format,
    )

    hsv = lp.fix_parameters(hsv, Nq=Nq)
    hsv = lp.prioritize_loops(hsv, "e,k,j,i")
    hsv = lp.tag_inames(hsv, {"e": "g.0", "j": "l.1", "i": "l.0"})
    hsv = lp.assume(hsv, "elements >= 1")

    hsv = fix_euler_parameters(hsv, p_p0=1, p_Gamma=1.4, p_R=1)
    from loopy.frontend.fortran.translator import specialize_fortran_division
    hsv = specialize_fortran_division(hsv)

    for name in ["Q", "rhsQ"]:
        hsv = set_q_storage_format(hsv, name)

    hsv = set_D_storage_format(hsv)
    # hsv = lp.add_prefetch(hsv, "volumeGeometricFactors")

    ref_hsv = hsv

    if opt_level == 0:
        tap_hsv = hsv

    hsv = lp.add_prefetch(hsv, "D[:,:]", fetch_outer_inames="e",
            default_tag="l.auto")

    if opt_level == 1:
        tap_hsv = hsv

    # turn the first reads into subst rules
    local_prep_var_names = set()
    for insn in lp.find_instructions(hsv, "tag:local_prep"):
        assignee, = insn.assignee_var_names()
        local_prep_var_names.add(assignee)
        hsv = lp.assignment_to_subst(hsv, assignee)

    # precompute fluxes
    hsv = lp.assignment_to_subst(hsv, "JinvD_r")
    hsv = lp.assignment_to_subst(hsv, "JinvD_s")

    r_fluxes = lp.find_instructions(hsv, "tag:compute_fluxes and tag:rknl")
    s_fluxes = lp.find_instructions(hsv, "tag:compute_fluxes and tag:sknl")

    if ilp_multiple > 1:
        hsv = lp.split_iname(hsv, "k", 2, inner_tag="ilp")
        ilp_inames = ("k_inner",)
        flux_ilp_inames = ("kk",)
    else:
        ilp_inames = ()
        flux_ilp_inames = ()

    rtmps = []
    stmps = []

    flux_store_idx = 0

    for rflux_insn, sflux_insn in zip(r_fluxes, s_fluxes):
        for knl_tag, insn, flux_inames, tmps, flux_precomp_inames in [
                  ("rknl", rflux_insn, ("j", "n",), rtmps, ("jj", "ii",)),
                  ("sknl", sflux_insn, ("i", "n",), stmps, ("ii", "jj",)),
                  ]:
            flux_var, = insn.assignee_var_names()
            print(insn)

            reader, = lp.find_instructions(hsv,
                  "tag:{knl_tag} and reads:{flux_var}"
                  .format(knl_tag=knl_tag, flux_var=flux_var))

            hsv = lp.assignment_to_subst(hsv, flux_var)

            flux_store_name = "flux_store_%d" % flux_store_idx
            flux_store_idx += 1
            tmps.append(flux_store_name)

            hsv = lp.precompute(hsv, flux_var+"_subst", flux_inames + ilp_inames,
                temporary_name=flux_store_name,
                precompute_inames=flux_precomp_inames + flux_ilp_inames,
                default_tag=None)
            if flux_var.endswith("_s"):
                hsv = lp.tag_array_axes(hsv, flux_store_name, "N0,N1,N2?")
            else:
                hsv = lp.tag_array_axes(hsv, flux_store_name, "N1,N0,N2?")

            n_iname = "n_"+flux_var.replace("_r", "").replace("_s", "")
            if n_iname.endswith("_0"):
                n_iname = n_iname[:-2]
            hsv = lp.rename_iname(hsv, "n", n_iname, within="id:"+reader.id,
                  existing_ok=True)

    hsv = lp.tag_inames(hsv, {"ii": "l.0", "jj": "l.1"})
    for iname in flux_ilp_inames:
        hsv = lp.tag_inames(hsv, {iname: "ilp"})

    hsv = lp.alias_temporaries(hsv, rtmps)
    hsv = lp.alias_temporaries(hsv, stmps)

    if opt_level == 2:
        tap_hsv = hsv

    for prep_var_name in local_prep_var_names:
        if prep_var_name.startswith("Jinv") or "_s" in prep_var_name:
            continue
        hsv = lp.precompute(hsv,
            lp.find_one_rule_matching(hsv, prep_var_name+"_*subst*"),
            default_tag="l.auto")

    if opt_level == 3:
        tap_hsv = hsv

    hsv = lp.add_prefetch(hsv, "Q[ii,jj,k,:,:,e]", sweep_inames=ilp_inames,
            default_tag="l.auto")

    if opt_level == 4:
        tap_hsv = hsv
        tap_hsv = lp.tag_inames(tap_hsv, {
              "Q_dim_field_inner": "unr",
              "Q_dim_field_outer": "unr"})

    hsv = lp.buffer_array(hsv, "rhsQ", ilp_inames,
          fetch_bounding_box=True, default_tag="for",
          init_expression="0", store_expression="base + buffer")

    if opt_level == 5:
        tap_hsv = hsv
        tap_hsv = lp.tag_inames(tap_hsv, {
              "rhsQ_init_field_inner": "unr", "rhsQ_store_field_inner": "unr",
              "rhsQ_init_field_outer": "unr", "rhsQ_store_field_outer": "unr",
              "Q_dim_field_inner": "unr",
              "Q_dim_field_outer": "unr"})

    # buffer axes need to be vectorized in order for this to work
    hsv = lp.tag_array_axes(hsv, "rhsQ_buf", "c?,vec,c")
    hsv = lp.tag_array_axes(hsv, "Q_fetch", "c?,vec,c")
    hsv = lp.tag_array_axes(hsv, "D_fetch", "f,f")
    hsv = lp.tag_inames(hsv,
            {"Q_dim_k": "unr", "rhsQ_init_k": "unr", "rhsQ_store_k": "unr"},
            ignore_nonexistent=True)

    if opt_level == 6:
        tap_hsv = hsv
        tap_hsv = lp.tag_inames(tap_hsv, {
              "rhsQ_init_field_inner": "unr", "rhsQ_store_field_inner": "unr",
              "rhsQ_init_field_outer": "unr", "rhsQ_store_field_outer": "unr",
              "Q_dim_field_inner": "unr",
              "Q_dim_field_outer": "unr"})

    hsv = lp.tag_inames(hsv, {
          "rhsQ_init_field_inner": "vec", "rhsQ_store_field_inner": "vec",
          "rhsQ_init_field_outer": "unr", "rhsQ_store_field_outer": "unr",
          "Q_dim_field_inner": "vec",
          "Q_dim_field_outer": "unr"})

    if opt_level == 7:
        tap_hsv = hsv

    hsv = lp.collect_common_factors_on_increment(hsv, "rhsQ_buf",
          vary_by_axes=(0,) if ilp_multiple > 1 else ())

    if opt_level >= 8:
        tap_hsv = hsv

    hsv = tap_hsv

    hsv = lp.preprocess_kernel(hsv)
    hsv = lp.allocate_temporaries_for_base_storage(hsv)

    hsv = lp.set_options(hsv,
            build_options=[
                 "-cl-denorms-are-zero",
                 "-cl-fast-relaxed-math",
                 "-cl-finite-math-only",
                 "-cl-mad-enable",
                 "-cl-no-signed-zeros"])

    if 1:
        print("OPS")
        op_map = lp.get_op_map(hsv, subgroup_size=32)
        print(op_map)

        print("MEM")
        gmem_map = lp.get_mem_access_map(hsv, subgroup_size=32).to_bytes()
        print(gmem_map)

    # FIXME: renaming's a bit tricky in this program model.
    # add a simple transformation for it
    # hsv = hsv.copy(name="horizontalStrongVolumeKernel")

    results = lp.auto_test_vs_ref(ref_hsv, ctx, hsv, parameters={"elements": 300},
            quiet=True)

    elapsed = results["elapsed_wall"]

    print("elapsed", elapsed)


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging

import numpy as np
import pytest

import pyopencl as cl  # noqa: F401
import pyopencl.clmath
import pyopencl.clrandom  # noqa: F401
from pyopencl.tools import (  # noqa: F401
    pytest_generate_tests_for_pyopencl as pytest_generate_tests,
)

import loopy as lp
from loopy.diagnostic import LoopyError
from loopy.version import LOOPY_USE_LANGUAGE_VERSION_2018_2  # noqa: F401


logger = logging.getLogger(__name__)


# {{{ convolutions

def test_convolution(ctx_factory):
    ctx = ctx_factory()

    dtype = np.float32

    knl = lp.make_kernel(
        "{ [iimg, ifeat, icolor, im_x, im_y, f_x, f_y]: \
                -f_w <= f_x,f_y <= f_w \
                and 0 <= im_x < im_w and 0 <= im_y < im_h \
                and 0<=iimg<=nimgs and 0<=ifeat<nfeats and 0<=icolor<ncolors \
                }",
        """
        out[iimg, ifeat, im_x, im_y] = sum((f_x, f_y, icolor), \
            img[iimg, f_w+im_x-f_x, f_w+im_y-f_y, icolor] \
            * f[ifeat, f_w+f_x, f_w+f_y, icolor])
        """,
        [
            lp.GlobalArg("f", dtype, shape=lp.auto),
            lp.GlobalArg("img", dtype, shape=lp.auto),
            lp.GlobalArg("out", dtype, shape=lp.auto),
            "..."
            ],
        assumptions="f_w>=1 and im_w, im_h >= 2*f_w+1 and nfeats>=1 and nimgs>=0",
        options="annotate_inames")

    f_w = 3

    knl = lp.fix_parameters(knl, f_w=f_w, ncolors=3)

    ref_knl = knl

    def variant_0(knl):
        # knl = lp.split_iname(knl, "im_x", 16, inner_tag="l.0")
        knl = lp.prioritize_loops(knl, "iimg,im_x,im_y,ifeat,f_x,f_y")
        return knl

    def variant_1(knl):
        knl = lp.split_iname(knl, "im_x", 16, inner_tag="l.0")
        knl = lp.prioritize_loops(knl, "iimg,im_x_outer,im_y,ifeat,f_x,f_y")
        return knl

    def variant_2(knl):
        knl = lp.split_iname(knl, "im_x", 16, outer_tag="g.0", inner_tag="l.0")
        knl = lp.split_iname(knl, "im_y", 16, outer_tag="g.1", inner_tag="l.1")
        knl = lp.tag_inames(knl, {"ifeat": "g.2"})
        knl = lp.add_prefetch(knl, "f[ifeat,:,:,:]",
                fetch_outer_inames="im_x_outer, im_y_outer, ifeat",
                default_tag="l.auto")
        knl = lp.add_prefetch(knl, "img", "im_x_inner, im_y_inner, f_x, f_y",
                fetch_outer_inames="iimg, im_x_outer, im_y_outer, ifeat, icolor",
                default_tag="l.auto")
        return knl

    for variant in [
            # variant_0,
            # variant_1,
            variant_2
            ]:
        lp.auto_test_vs_ref(ref_knl, ctx, variant(knl),
                parameters={
                    "im_w": 128, "im_h": 128, "f_w": f_w,
                    "nfeats": 3, "nimgs": 3
                    })


def test_convolution_with_nonzero_base(ctx_factory):
    # This is kept alive as a test for domains that don't start at zero.
    # These are a bad idea for split_iname, which places its origin at zero
    # and therefore produces a first block that is odd-sized.
    #
    # Therefore, for real tests, check test_convolution further up.

    ctx = ctx_factory()

    dtype = np.float32

    knl = lp.make_kernel(
        "{ [iimg, ifeat, icolor, im_x, im_y, f_x, f_y]: \
                -f_w <= f_x,f_y <= f_w \
                and f_w <= im_x < im_w-f_w and f_w <= im_y < im_h-f_w \
                and 0<=iimg<=nimgs and 0<=ifeat<nfeats and 0<=icolor<ncolors \
                }",
        """
        out[iimg, ifeat, im_x-f_w, im_y-f_w] = sum((f_x, f_y, icolor), \
            img[iimg, im_x-f_x, im_y-f_y, icolor] \
            * f[ifeat, f_w+f_x, f_w+f_y, icolor])
        """,
        [
            lp.GlobalArg("f", dtype, shape=lp.auto),
            lp.GlobalArg("img", dtype, shape=lp.auto),
            lp.GlobalArg("out", dtype, shape=lp.auto),
            "..."
            ],
        assumptions="f_w>=1 and im_w, im_h >= 2*f_w+1 and nfeats>=1 and nimgs>=0",
        options="annotate_inames")

    knl = lp.fix_parameters(knl, ncolors=3)

    ref_knl = knl

    f_w = 3

    def variant_0(knl):
        # knl = lp.split_iname(knl, "im_x", 16, inner_tag="l.0")
        knl = lp.prioritize_loops(knl, "iimg,im_x,im_y,ifeat,f_x,f_y")
        return knl

    def variant_1(knl):
        knl = lp.split_iname(knl, "im_x", 16, inner_tag="l.0")
        knl = lp.prioritize_loops(knl, "iimg,im_x_outer,im_y,ifeat,f_x,f_y")
        return knl

    for variant in [
            variant_0,
            variant_1,
            ]:
        lp.auto_test_vs_ref(ref_knl, ctx, variant(knl),
                parameters={
                    "im_w": 128, "im_h": 128, "f_w": f_w,
                    "nfeats": 12, "nimgs": 17
                    })

# }}}


def test_rob_stroud_bernstein():
    # NOTE: tmp would have to be zero-filled beforehand

    knl = lp.make_kernel(
            "{[el, i2, alpha1,alpha2]: \
                    0 <= el < nels and \
                    0 <= i2 < nqp1d and \
                    0 <= alpha1 <= deg and 0 <= alpha2 <= deg-alpha1 }",
            """
            for el,i2
                <> xi = qpts[1, i2]
                <> s = 1-xi
                <> r = xi/s
                <> aind = 0 {id=aind_init}

                for alpha1
                    <> w = s**(deg-alpha1) {id=init_w}

                    for alpha2
                        tmp[el,alpha1,i2] = tmp[el,alpha1,i2] + w * coeffs[aind] \
                                {id=write_tmp,dep=init_w:aind_init}
                        w = w * r * ( deg - alpha1 - alpha2 ) / (1 + alpha2) \
                                {id=update_w,dep=init_w:write_tmp}
                        aind = aind + 1 \
                                {id=aind_incr,dep=aind_init:write_tmp:update_w}
                    end
                end
            end
            """,
            [
                # Must declare coeffs to have "no" shape, to keep loopy
                # from trying to figure it out the shape automatically.

                lp.GlobalArg("coeffs", None, shape=None),
                "..."
                ],
            assumptions="deg>=0 and nels>=1",
            target=lp.PyOpenCLTarget()
            )

    knl = lp.fix_parameters(knl, nqp1d=7, deg=4)
    knl = lp.split_iname(knl, "el", 16, inner_tag="l.0")
    knl = lp.split_iname(knl, "el_outer", 2, outer_tag="g.0", inner_tag="ilp",
            slabs=(0, 1))
    knl = lp.tag_inames(knl, {"i2": "l.1", "alpha1": "unr", "alpha2": "unr"})
    knl = lp.add_dtypes(knl, {
                "qpts": np.float32,
                "coeffs": np.float32,
                "tmp": np.float32,
                })
    print(lp.generate_code_v2(knl))


def test_rob_stroud_bernstein_full():
    # NOTE: result would have to be zero-filled beforehand

    knl = lp.make_kernel(
        "{[el, i2, alpha1,alpha2, i1_2, alpha1_2, i2_2]: \
                0 <= el < nels and \
                0 <= i2 < nqp1d and \
                0 <= alpha1 <= deg and 0 <= alpha2 <= deg-alpha1 and\
                \
                0 <= i1_2 < nqp1d and \
                0 <= alpha1_2 <= deg and \
                0 <= i2_2 < nqp1d \
                }",
        """
        for el
            for i2
                <> xi = qpts[1, i2]
                <> s = 1-xi
                <> r = xi/s
                <> aind = 0 {id=aind_init}

                for alpha1
                    <> w = s**(deg-alpha1) {id=init_w}

                    <> tmp[alpha1,i2] = tmp[alpha1,i2] + w * coeffs[aind] \
                            {id=write_tmp,dep=init_w:aind_init}
                    for alpha2
                        w = w * r * ( deg - alpha1 - alpha2 ) / (1 + alpha2) \
                            {id=update_w,dep=init_w:write_tmp}
                        aind = aind + 1 \
                            {id=aind_incr,dep=aind_init:write_tmp:update_w}
                    end
                end
            end

            for i1_2
                <> xi2 = qpts[0, i1_2] {dep=aind_incr}
                <> s2 = 1-xi2
                <> r2 = xi2/s2
                <> w2 = s2**deg  {id=w2_init}

                for alpha1_2
                    for i2_2
                        result[el, i1_2, i2_2] = result[el, i1_2, i2_2] + \
                                w2 * tmp[alpha1_2, i2_2]  {id=res2,dep=w2_init}
                    end

                    w2 = w2 * r2 * (deg-alpha1_2) / (1+alpha1_2)  \
                            {id=w2_update, dep=res2}
                end
            end
        end
        """,
        [
            # Must declare coeffs to have "no" shape, to keep loopy
            # from trying to figure it out the shape automatically.

            lp.GlobalArg("coeffs", None, shape=None),
            "..."
            ],
        assumptions="deg>=0 and nels>=1",
        target=lp.PyOpenCLTarget()
        )

    knl = lp.fix_parameters(knl, nqp1d=7, deg=4)

    if 0:
        knl = lp.split_iname(knl, "el", 16, inner_tag="l.0")
        knl = lp.split_iname(knl, "el_outer", 2, outer_tag="g.0", inner_tag="ilp",
                slabs=(0, 1))
        knl = lp.tag_inames(knl, {"i2": "l.1", "alpha1": "unr", "alpha2": "unr"})

    from pickle import dumps, loads
    knl = loads(dumps(knl))

    knl = lp.add_dtypes(knl,
            {
                "qpts": np.float32,
                "tmp": np.float32,
                "coeffs": np.float32,
                "result": np.float32,
                })
    print(lp.generate_code_v2(knl))


def test_stencil(ctx_factory):
    ctx = ctx_factory()

    # n=32 causes corner case behavior in size calculations for temporary (a
    # non-unifiable, two-constant-segments PwAff as the base index)

    n = 256
    knl = lp.make_kernel(
            "{[i,j]: 0<= i,j < %d}" % n,
            [
                "a_offset(ii, jj) := a[ii+1, jj+1]",
                "z[i,j] = -2*a_offset(i,j)"
                " + a_offset(i,j-1)"
                " + a_offset(i,j+1)"
                " + a_offset(i-1,j)"
                " + a_offset(i+1,j)"
                ],
            [
                lp.GlobalArg("a", np.float32, shape=(n+2, n+2,)),
                lp.GlobalArg("z", np.float32, shape=(n+2, n+2,))
                ])

    ref_knl = knl

    def variant_1(knl):
        knl = lp.split_iname(knl, "i", 16, outer_tag="g.1", inner_tag="l.1")
        knl = lp.split_iname(knl, "j", 16, outer_tag="g.0", inner_tag="l.0")
        knl = lp.add_prefetch(knl, "a", ["i_inner", "j_inner"], default_tag="l.auto")
        knl = lp.prioritize_loops(knl, ["a_dim_0_outer", "a_dim_1_outer"])
        return knl

    def variant_2(knl):
        knl = lp.split_iname(knl, "i", 16, outer_tag="g.1", inner_tag="l.1")
        knl = lp.split_iname(knl, "j", 16, outer_tag="g.0", inner_tag="l.0")
        knl = lp.add_prefetch(knl, "a", ["i_inner", "j_inner"],
                fetch_bounding_box=True, default_tag="l.auto")
        knl = lp.prioritize_loops(knl, ["a_dim_0_outer", "a_dim_1_outer"])
        return knl

    for variant in [
            # variant_1,
            variant_2,
            ]:
        lp.auto_test_vs_ref(ref_knl, ctx, variant(knl),
                print_ref_code=False,
                op_count=[n*n], op_label=["cells"])


def test_stencil_with_overfetch(ctx_factory):
    ctx = ctx_factory()

    knl = lp.make_kernel(
            "{[i,j]: 0<= i,j < n}",
            [
                "a_offset(ii, jj) := a[ii+2, jj+2]",
                "z[i,j] = -2*a_offset(i,j)"
                " + a_offset(i,j-1)"
                " + a_offset(i,j+1)"
                " + a_offset(i-1,j)"
                " + a_offset(i+1,j)"

                " + a_offset(i,j-2)"
                " + a_offset(i,j+2)"
                " + a_offset(i-2,j)"
                " + a_offset(i+2,j)"
                ],
            assumptions="n>=1")

    if ctx.devices[0].platform.name == "Portable Computing Language":
        # https://github.com/pocl/pocl/issues/205
        pytest.skip("takes very long to compile on pocl")

    knl = lp.add_and_infer_dtypes(knl, {"a": np.float32})

    ref_knl = knl

    def variant_overfetch(knl):
        knl = lp.split_iname(knl, "i", 16, outer_tag="g.1", inner_tag="l.1",
                slabs=(1, 1))
        knl = lp.split_iname(knl, "j", 16, outer_tag="g.0", inner_tag="l.0",
               slabs=(1, 1))
        knl = lp.add_prefetch(knl, "a", ["i_inner", "j_inner"],
                fetch_bounding_box=True, default_tag="l.auto")
        knl = lp.prioritize_loops(knl, ["a_dim_0_outer", "a_dim_1_outer"])
        return knl

    for variant in [variant_overfetch]:
        n = 200
        lp.auto_test_vs_ref(ref_knl, ctx, variant(knl),
                print_ref_code=False,
                op_count=[n*n], parameters={"n": n}, op_label=["cells"])


def test_sum_factorization():
    knl = lp.make_kernel(
        "{[i,j,ip,jp,k,l]: "
        "0<=i<I and 0<=j<J and 0<=ip<IP and 0<=jp<JP and 0<=k,l<Q}",
        """
        phi1(i, x) := x**i
        phi2(i, x) := x**i
        psi1(i, x) := x**i
        psi2(i, x) := x**i
        a(x, y) := 1

        A[i,j,ip,jp] = sum(k,sum(l,
            phi1(i,x[0,k]) * phi2(j,x[1,l])
            * psi1(ip, x[0,k]) * psi2(jp, x[1, l])
            * w[0,k] * w[1,l]
            * a(x[0,k], x[1,l])
        ))
        """)

    pytest.xfail("extract_subst is currently too stupid for sum factorization")

    knl = lp.extract_subst(knl, "temp_array",
            "phi1(i,x[0,k]) *psi1(ip, x[0,k]) * w[0,k]")
    knl = lp.extract_subst(knl, "temp_array",
            "sum(k, phi1(i,x[0,k]) *psi1(ip, x[0,k]) * w[0,k])")

    print(knl)


def test_lbm(ctx_factory):
    ctx = ctx_factory()

    # D2Q4Q4Q4 lattice Boltzmann scheme for the shallow water equations
    # Example by Loic Gouarin <loic.gouarin@math.u-psud.fr>
    knl = lp.make_kernel(
        "{[ii,jj]:0<=ii<nx-2 and 0<=jj<ny-2}",
        """
        i := ii + 1
        j := jj + 1
        for ii, jj
            with {id_prefix=init_m}
                <> m[0] =   +    f[i-1, j, 0] +    f[i, j-1, 1] + f[i+1, j, 2] +  f[i, j+1, 3]
                m[1] =   + 4.*f[i-1, j, 0] - 4.*f[i+1, j, 2]
                m[2] =   + 4.*f[i, j-1, 1] - 4.*f[i, j+1, 3]
                m[3] =   +    f[i-1, j, 0] -    f[i, j-1, 1] + f[i+1, j, 2] -  f[i, j+1, 3]
                m[4] =   +    f[i-1, j, 4] +    f[i, j-1, 5] + f[i+1, j, 6] +  f[i, j+1, 7]
                m[5] =   + 4.*f[i-1, j, 4] - 4.*f[i+1, j, 6]
                m[6] =   + 4.*f[i, j-1, 5] - 4.*f[i, j+1, 7]
                m[7] =   +    f[i-1, j, 4] -    f[i, j-1, 5] + f[i+1, j, 6] -  f[i, j+1, 7]
                m[8] =   +    f[i-1, j, 8] +    f[i, j-1, 9] + f[i+1, j, 10] + f[i, j+1, 11]
                m[9] =   + 4.*f[i-1, j, 8] - 4.*f[i+1, j, 10]
                m[10] =  + 4.*f[i, j-1, 9] - 4.*f[i, j+1, 11]
                m[11] =  +    f[i-1, j, 8] -    f[i, j-1, 9] + f[i+1, j, 10] - f[i, j+1, 11]
            end

            with {id_prefix=update_m,dep=init_m*}
                m[1] = m[1] + 2.*(m[4] - m[1])
                m[2] = m[2] + 2.*(m[8] - m[2])
                m[3] = m[3]*(1. - 1.5)
                m[5] = m[5] + 1.5*(0.5*(m[0]*m[0]) + (m[4]*m[4])/m[0] - m[5])
                m[6] = m[6] + 1.5*(m[4]*m[8]/m[0] - m[6])
                m[7] = m[7]*(1. - 1.2000000000000000)
                m[9] = m[9] + 1.5*(m[4]*m[8]/m[0] - m[9])
                m[10] = m[10] + 1.5*(0.5*(m[0]*m[0]) + (m[8]*m[8])/m[0] - m[10])
                m[11] = m[11]*(1. - 1.2)
            end

            with {dep=update_m*}
                f_new[i, j, 0] =  + 0.25*m[0] + 0.125*m[1] + 0.25*m[3]
                f_new[i, j, 1] =  + 0.25*m[0] + 0.125*m[2] - 0.25*m[3]
                f_new[i, j, 2] =  + 0.25*m[0] - 0.125*m[1] + 0.25*m[3]
                f_new[i, j, 3] =  + 0.25*m[0] - 0.125*m[2] - 0.25*m[3]
                f_new[i, j, 4] =  + 0.25*m[4] + 0.125*m[5] + 0.25*m[7]
                f_new[i, j, 5] =  + 0.25*m[4] + 0.125*m[6] - 0.25*m[7]
                f_new[i, j, 6] =  + 0.25*m[4] - 0.125*m[5] + 0.25*m[7]
                f_new[i, j, 7] =  + 0.25*m[4] - 0.125*m[6] - 0.25*m[7]
                f_new[i, j, 8] =  + 0.25*m[8] + 0.125*m[9] + 0.25*m[11]
                f_new[i, j, 9] =  + 0.25*m[8] + 0.125*m[10] - 0.25*m[11]
                f_new[i, j, 10] =  + 0.25*m[8] - 0.125*m[9] + 0.25*m[11]
                f_new[i, j, 11] =  + 0.25*m[8] - 0.125*m[10] - 0.25*m[11]
           end
        end
        """)  # noqa: E501

    knl = lp.add_and_infer_dtypes(knl, {"f": np.float32})

    ref_knl = knl

    knl = lp.split_iname(knl, "ii", 16, outer_tag="g.1", inner_tag="l.1")
    knl = lp.split_iname(knl, "jj", 16, outer_tag="g.0", inner_tag="l.0")
    knl = lp.expand_subst(knl)
    knl = lp.add_prefetch(knl, "f", "ii_inner,jj_inner", fetch_bounding_box=True,
            default_tag="l.auto")

    lp.auto_test_vs_ref(ref_knl, ctx, knl, parameters={"nx": 20, "ny": 20})


def test_fd_demo():
    knl = lp.make_kernel(
        "{[i,j]: 0<=i,j<n}",
        "result[i+1,j+1] = u[i + 1, j + 1]**2 + -1 + (-4)*u[i + 1, j + 1] \
                + u[i + 1 + 1, j + 1] + u[i + 1 + -1, j + 1] \
                + u[i + 1, j + 1 + 1] + u[i + 1, j + 1 + -1]")
    # assumptions="n mod 16=0")
    knl = lp.split_iname(knl,
            "i", 16, outer_tag="g.1", inner_tag="l.1")
    knl = lp.split_iname(knl,
            "j", 16, outer_tag="g.0", inner_tag="l.0")
    knl = lp.add_prefetch(knl, "u",
            ["i_inner", "j_inner"],
            fetch_bounding_box=True,
            default_tag="l.auto")

    # n = 1000
    # u = cl.clrandom.rand(queue, (n+2, n+2), dtype=np.float32)

    knl = lp.set_options(knl, write_code=True)
    knl = lp.add_and_infer_dtypes(knl, {"u": np.float32})
    code, _inf = lp.generate_code(knl)
    print(code)

    assert "double" not in code


def test_fd_1d(ctx_factory):
    ctx = ctx_factory()

    knl = lp.make_kernel(
        "{[i]: 0<=i<n}",
        "result[i] = u[i+1]-u[i]")

    knl = lp.add_and_infer_dtypes(knl, {"u": np.float32})
    ref_knl = knl

    knl = lp.split_iname(knl, "i", 16)
    knl = lp.extract_subst(knl, "u_acc", "u[j]", parameters="j")
    knl = lp.precompute(knl, "u_acc", "i_inner", default_tag="for")
    knl = lp.assume(knl, "n mod 16 = 0")

    lp.auto_test_vs_ref(
            ref_knl, ctx, knl,
            parameters={"n": 2048})


def test_poisson_fem(ctx_factory):
    # Stolen from Peter Coogan and Rob Kirby for FEM assembly
    ctx = ctx_factory()

    nbf = 5
    nqp = 5
    sdim = 3

    knl = lp.make_kernel(
            "{ [c,i,j,k,ell,ell2]: \
            0 <= c < nels and \
            0 <= i < nbf and \
            0 <= j < nbf and \
            0 <= k < nqp and \
            0 <= ell,ell2 < sdim}",
            """
            dpsi(bf,k0,dir) := \
                    simul_reduce(sum, ell2, DFinv[c,ell2,dir] * DPsi[bf,k0,ell2] )
            Ael[c,i,j] = \
                    J[c] * w[k] * sum(ell, dpsi(i,k,ell) * dpsi(j,k,ell))
            """,
            assumptions="nels>=1 and nbf >= 1 and nels mod 4 = 0")

    print(knl)

    knl = lp.fix_parameters(knl, nbf=nbf, sdim=sdim, nqp=nqp)

    ref_knl = knl

    knl = lp.prioritize_loops(knl, ["c", "j", "i", "k"])

    def variant_1(knl):
        knl = lp.precompute(knl, "dpsi", "i,k,ell", default_tag="for")
        knl = lp.prioritize_loops(knl, "c,i,j")
        return knl

    def variant_2(knl):
        knl = lp.precompute(knl, "dpsi", "i,ell", default_tag="for")
        knl = lp.prioritize_loops(knl, "c,i,j")
        return knl

    def add_types(knl):
        return lp.add_and_infer_dtypes(knl, {
            "w": np.float32,
            "J": np.float32,
            "DPsi": np.float32,
            "DFinv": np.float32,
            })

    for variant in [
            # variant_1,
            variant_2
            ]:
        knl = variant(knl)

        lp.auto_test_vs_ref(
                add_types(ref_knl), ctx, add_types(knl),
                parameters={"n": 5, "nels": 15, "nbf": 5, "sdim": 2, "nqp": 7})


def test_domain_tree_nesting():
    # From https://github.com/inducer/loopy/issues/78

    AS = lp.AddressSpace        # noqa: N806

    out_map = np.array([1, 2], dtype=np.int32)
    if_val = np.array([-1, 0], dtype=np.int32)
    vals = np.array([2, 3], dtype=np.int32)
    num_vals = np.array([2, 4], dtype=np.int32)
    num_vals_offset = np.array(np.cumsum(num_vals) - num_vals, dtype=np.int32)

    TV = lp.TemporaryVariable  # noqa: N806

    knl = lp.make_kernel(["{[i]: 0 <= i < 12}",
                    "{[j]: 0 <= j < 100}",
                    "{[a_count]: 0 <= a_count < a_end}",
                    "{[b_count]: 0 <= b_count < b_end}"],
    """
    for j
        for i
            <> a_end = abs(if_val[i])

            <>b_end = num_vals[i]
            <>offset = num_vals_offset[i] {id=offset}
            <>b_sum = 0 {id=b_init}
            for b_count
                <>val = vals[offset + b_count] {dep=offset}
            end
            b_sum = exp(b_sum) {id=b_final}

            out[j,i] =  b_sum {dep=b_final}
        end
    end
    """,
    [
        TV("out_map", initializer=out_map, read_only=True, address_space=AS.PRIVATE),
        TV("if_val", initializer=if_val, read_only=True, address_space=AS.PRIVATE),
        TV("vals", initializer=vals, read_only=True, address_space=AS.PRIVATE),
        TV("num_vals", initializer=num_vals, read_only=True,
           address_space=AS.PRIVATE),
        TV("num_vals_offset", initializer=num_vals_offset, read_only=True,
           address_space=AS.PRIVATE),
        lp.GlobalArg("B", shape=(100, 31), dtype=np.float64),
        lp.GlobalArg("out", shape=(100, 12), dtype=np.float64)],
        name="nested_domain")

    parents_per_domain = knl["nested_domain"].parents_per_domain()

    def depth(i):
        if parents_per_domain[i] is None:
            return 0
        else:
            return 1 + depth(parents_per_domain[i])

    for i in range(len(parents_per_domain)):
        assert depth(i) < 2


def test_prefetch_through_indirect_access():
    knl = lp.make_kernel("{[i, j, k]: 0 <= i,k < 10 and 0<=j<2}",
        """
        for i, j, k
            a[map1[indirect[i], j], k] = 2
        end
        """,
        [
            lp.GlobalArg("a", strides=(2, 1), dtype=int),
            lp.GlobalArg("map1", shape=(10, 10), dtype=int),
            "..."
            ],
        target=lp.CTarget())

    knl = lp.prioritize_loops(knl, "i,j,k")

    with pytest.raises(LoopyError):
        knl = lp.add_prefetch(knl, "map1[:, j]", default_tag="l.auto")


def test_unsigned_types_to_mod():
    knl = lp.make_kernel("{[i]: 0<=i<10}",
        """
            <> c = b[i] {id=init,dup=i}
            a[i] = i % c {dep=init}
        """,
        [lp.GlobalArg("a", shape=(10,), dtype=np.uint32),
         lp.GlobalArg("b", shape=(10,), dtype=np.uint32)]
    )
    assert "loopy_mod" not in lp.generate_code_v2(knl).device_code()


def test_abs_as_index():
    knl = lp.make_kernel(
        ["{[i]: 0<=i<10}"],
        """
        b[i] = a[abs(5-i)]
        """,
        [
            lp.GlobalArg("a", np.float32),
            ...
            ])
    print(lp.generate_code_v2(knl).device_code())


def test_sumpy_p2p_reduced():
    knl = lp.make_kernel(
        [
            "{[itgt_box]: 0<=itgt_box<5 }",
            "{[isrc_box]: 0<=isrc_box<isrc_box_end }",
            "{[inner]: 0 <=inner<=31 }",
            "{[itgt_offset_outer]: itgt_offset_outer=0 }",
            "{[isrc_prefetch_inner]: isrc_prefetch_inner=0 and 0<=inner<=25 }",
        ],
        """
        <> isrc_box_end = source_box_starts[itgt_box + 1] {inames=inner:itgt_box}
        <> itgt_offset = inner {inames=inner:itgt_box}
        <> isrc_prefetch = isrc_prefetch_inner*32 + inner \
            {inames=isrc_prefetch_inner:inner:isrc_box:itgt_box}
        """,
        [
            lp.GlobalArg("box_target_starts", dtype=np.int32),
            lp.GlobalArg("box_target_counts_nonchild", dtype=np.int32),
            lp.GlobalArg("box_source_starts", dtype=np.int32),
            lp.GlobalArg("box_source_counts_nonchild", dtype=np.int32),
            lp.GlobalArg("source_box_starts", dtype=np.int32),
            lp.GlobalArg("source_box_lists", dtype=np.int32),
        ],
        silenced_warnings="unused_inames",
    )
    lp.generate_code_v2(knl).device_code()


if __name__ == "__main__":
    import sys
    if len(sys.argv) > 1:
        exec(sys.argv[1])
    else:
        from pytest import main
        main([__file__])

# vim: foldmethod=marker



import os
from urllib.request import urlopen


_conf_url = "https://raw.githubusercontent.com/inducer/sphinxconfig/main/sphinxconfig.py"
with urlopen(_conf_url) as _inf:
    exec(compile(_inf.read(), _conf_url, "exec"), globals())

copyright = "2016, Andreas Klöckner"

# The short X.Y version.
ver_dic = {}
_version_source = "../loopy/version.py"
with open(_version_source) as vpy_file:
    version_py = vpy_file.read()

os.environ["AKPYTHON_EXEC_IMPORT_UNAVAILABLE"] = "1"
exec(compile(version_py, _version_source, "exec"), ver_dic)
version = ".".join(str(x) for x in ver_dic["VERSION"])
# The full version, including alpha/beta/rc tags.
release = ver_dic["VERSION_TEXT"]
del os.environ["AKPYTHON_EXEC_IMPORT_UNAVAILABLE"]

exclude_patterns = ["_build"]

# Example configuration for intersphinx: refer to the Python standard library.
intersphinx_mapping = {
        "python": ("https://docs.python.org/3", None),
        "numpy": ("https://numpy.org/doc/stable/", None),
        "pytools": ("https://documen.tician.de/pytools", None),
        "islpy": ("https://documen.tician.de/islpy", None),
        "pyopencl": ("https://documen.tician.de/pyopencl", None),
        "cgen": ("https://documen.tician.de/cgen", None),
        "pymbolic": ("https://documen.tician.de/pymbolic", None),
        "constantdict": ("https://matthiasdiener.github.io/constantdict/", None),
        }

nitpicky = True

nitpick_ignore_regex = [
        ["py:class", r"typing_extensions\.(.+)"],
        ["py:class", r"numpy\.u?int[0-9]+"],
        ["py:class", r"numpy\.float[0-9]+"],
        ["py:class", r"numpy\.complex[0-9]+"],

        # Reference not found from "<unknown>"? I'm not even sure where to look.
        ["py:class", r"ExpressionNode"],

        # Type aliases
        ["py:class", r"InameStr"],
        ["py:class", r"ConcreteCallablesTable"],
        ["py:class", r"LoopNestTree"],
        ["py:class", r"LoopTree"],
        ["py:class", r"ToLoopyTypeConvertible"],
        ["py:class", r"ToStackMatchConvertible"],
        ]



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

import logging
from collections import defaultdict
from functools import reduce
from typing import TYPE_CHECKING

import numpy as np

import islpy as isl
from islpy import dim_type
from pymbolic.primitives import Variable, is_arithmetic_expression
from pytools import memoize_method

from loopy.diagnostic import (
    LoopyError,
    LoopyIndexError,
    WriteRaceConditionWarning,
    warn_with_kernel,
)
from loopy.kernel.array import (
    ArrayBase,
    FixedStrideArrayDimTag,
    SeparateArrayArrayDimTag,
)
from loopy.kernel.data import (
    AddressSpace,
    ArrayArg,
    ArrayDimImplementationTag,
    AxisTag,
    InameImplementationTag,
    TemporaryVariable,
    auto,
)
from loopy.kernel.function_interface import CallableKernel
from loopy.kernel.instruction import (
    CallInstruction,
    CInstruction,
    MultiAssignmentBase,
    NoOpInstruction,
    _DataObliviousInstruction,
)
from loopy.symbolic import CombineMapper, ResolvedFunction, SubArrayRef, WalkMapper
from loopy.translation_unit import (
    CallablesTable,
    TranslationUnit,
    check_each_kernel,
)
from loopy.type_inference import TypeReader
from loopy.typing import not_none


if TYPE_CHECKING:
    from collections.abc import Mapping, Sequence

    from pymbolic.typing import Expression

    from loopy.kernel import LoopKernel


logger = logging.getLogger(__name__)


__doc__ = """
.. currentmodule:: loopy.check

.. autofunction:: check_for_integer_subscript_indices

.. autofunction:: check_for_duplicate_insn_ids

.. autofunction:: check_for_double_use_of_hw_axes

.. autofunction:: check_insn_attributes

.. autofunction:: check_loop_priority_inames_known

.. autofunction:: check_multiple_tags_allowed

.. autofunction:: check_for_inactive_iname_access

.. autofunction:: check_for_unused_inames

.. autofunction:: check_for_write_races

.. autofunction:: check_for_data_dependent_parallel_bounds

.. autofunction:: check_bounds

.. autofunction:: check_variable_access_ordered
"""


# {{{ sanity checks run before preprocessing

def check_identifiers_in_subst_rules(knl):
    """Substitution rules may only refer to kernel-global quantities or their
    own arguments.
    """

    from loopy.symbolic import get_dependencies

    allowed_identifiers = knl.all_variable_names()

    for rule in knl.substitutions.values():
        deps = get_dependencies(rule.expression)
        rule_allowed_identifiers = allowed_identifiers | frozenset(rule.arguments)

        if not deps <= rule_allowed_identifiers:
            raise LoopyError("kernel '%s': substitution rule '%s' refers to "
                    "identifier(s) '%s' which are neither rule arguments nor "
                    "kernel-global identifiers"
                    % (knl.name, rule.name,
                       ", ".join(deps-rule_allowed_identifiers)))


class UnresolvedCallCollector(CombineMapper):
    """
    Collects all the unresolved calls within a kernel.

    :returns:
        A :class:`frozenset` of function names that are not resolved.
    """

    def combine(self, values):
        import operator
        return reduce(operator.or_, values, frozenset())

    def map_call(self, expr):
        if not isinstance(expr.function, ResolvedFunction):
            return frozenset([expr.function.name]) | self.rec(expr.parameters)
        else:
            return self.rec(expr.parameters)

    def map_call_with_kwargs(self, expr):
        # See: https://github.com/inducer/loopy/pull/323
        raise NotImplementedError

    def map_constant(self, expr):
        return frozenset()

    map_variable = map_constant
    map_function_symbol = map_constant
    map_tagged_variable = map_constant
    map_type_cast = map_constant
    map_nan = map_constant


@check_each_kernel
def check_functions_are_resolved(kernel: LoopKernel) -> None:
    """ Checks if all call nodes in the *kernel* expression have been
    resolved.
    """
    from loopy.symbolic import SubstitutionRuleExpander
    subst_expander = SubstitutionRuleExpander(kernel.substitutions)

    for insn in kernel.instructions:
        if isinstance(insn, MultiAssignmentBase):
            unresolved_calls = UnresolvedCallCollector()(subst_expander(insn
                                                                        .expression))
            if unresolved_calls:
                raise LoopyError("Unknown function '%s' -- register a "
                                 "callable corresponding to it." %
                                 set(unresolved_calls).pop())
        elif isinstance(insn, (CInstruction, _DataObliviousInstruction)):
            pass
        else:
            raise NotImplementedError(type(insn))


@check_each_kernel
def check_separated_array_consistency(kernel: LoopKernel) -> None:
    # Boo. This is (part of) the price of redundant representation.
    for arg in kernel.args:
        if isinstance(arg, ArrayArg) and arg._separation_info is not None:
            sep_indices = arg._separation_info.sep_axis_indices_set
            for subarg_name in arg._separation_info.subarray_names.values():
                sub_arg = kernel.arg_dict[subarg_name]

                from loopy.preprocess import _remove_at_indices

                assert arg.shape is None or isinstance(arg.shape, tuple)
                if _remove_at_indices(sep_indices, arg.shape) != sub_arg.shape:
                    raise LoopyError(
                            f"Shapes of '{arg.name}' and associated sep array "
                            "'{sub_arg.name}' are not consistent.")

                assert arg.dim_tags is None or isinstance(arg.dim_tags, tuple)
                if _remove_at_indices(sep_indices, arg.dim_tags) != sub_arg.dim_tags:
                    raise LoopyError(
                            f"Dim tags of '{arg.name}' and associated sep array "
                            "'{sub_arg.name}' are not consistent.")

                for attr_name in ["address_space", "is_input", "is_output"]:
                    if getattr(arg, attr_name) != getattr(sub_arg, attr_name):
                        raise LoopyError(
                                f"Attribute '{attr_name}' of "
                                f"'{arg.name}' and associated sep array "
                                f"'{sub_arg.name}' is not consistent.")


@check_each_kernel
def check_offsets_and_dim_tags(kernel: LoopKernel) -> None:
    from pymbolic.primitives import ExpressionNode, Variable

    from loopy.symbolic import DependencyMapper

    arg_name_vars = {Variable(name) for name in kernel.arg_dict}
    dep_mapper: DependencyMapper[[]] = DependencyMapper()

    def ensure_depends_only_on_arguments(
            what: str, expr: str | Expression) -> None:
        if isinstance(expr, str):
            expr = Variable(expr)

        deps = dep_mapper(expr)
        if not deps <= arg_name_vars:
            raise LoopyError(
                    f"not all names in {what} are arguments: "
                    + ", ".join(str(v) for v in deps - arg_name_vars))

    # {{{ process arguments

    new_args = []
    for arg in kernel.args:
        if isinstance(arg, ArrayArg):
            what = f"offset of argument '{arg.name}'"
            if arg.offset is None:
                continue
            if arg.offset is auto:
                pass
            elif isinstance(arg.offset, (int, np.integer, ExpressionNode, str)):
                ensure_depends_only_on_arguments(what, arg.offset)

            else:
                raise LoopyError(f"invalid value of offset for '{arg.name}'")

            if arg.dim_tags is None:
                new_dim_tags: tuple[ArrayDimImplementationTag, ...] | None = \
                        arg.dim_tags
            else:
                new_dim_tags = ()
                for iaxis, dim_tag in enumerate(arg.dim_tags):
                    if isinstance(dim_tag, FixedStrideArrayDimTag):
                        what = (f"stride for axis {iaxis+1} (1-based) of "
                                        f"of argument '{arg.name}'")
                        if dim_tag.stride is auto:
                            pass
                        elif isinstance(
                                dim_tag.stride, (int, np.integer, ExpressionNode)):
                            ensure_depends_only_on_arguments(what, dim_tag.stride)
                        else:
                            raise LoopyError(f"invalid value of {what}")

                    assert new_dim_tags is not None
                    new_dim_tags = (*new_dim_tags, dim_tag)

            arg = arg.copy(dim_tags=new_dim_tags)

        new_args.append(arg)

    # }}}

    # {{{ process temporary variables

    for tv in kernel.temporary_variables.values():
        what = f"offset of temporary '{tv.name}'"
        if tv.offset is None:
            pass
        if tv.offset is auto:
            pass
        elif isinstance(tv.offset, (int, np.integer, ExpressionNode, str)):
            ensure_depends_only_on_arguments(what, tv.offset)
        else:
            raise LoopyError(f"invalid value of offset for '{tv.name}'")

        if tv.dim_tags is not None:
            for iaxis, dim_tag in enumerate(tv.dim_tags):
                if isinstance(dim_tag, FixedStrideArrayDimTag):
                    what = ("axis stride for axis "
                            f"{iaxis+1} (1-based) of temporary '{tv.name}'")
                    if dim_tag.stride is auto:
                        raise LoopyError(f"The {what}" f" is 'auto', "
                                "which is not allowed.")
                    elif isinstance(dim_tag.stride, (int, np.integer, ExpressionNode)):
                        ensure_depends_only_on_arguments(what, dim_tag.stride)
                    else:
                        raise LoopyError(f"invalid value of {what}")

                elif isinstance(dim_tag, SeparateArrayArrayDimTag):
                    raise LoopyError(f"Axis {iaxis+1} of temporary "
                            f"'{tv.name} is tagged 'sep'. This is not allowed.")

    # }}}

# }}}


# {{{ sanity checks run pre-scheduling

# FIXME: Replace with an enum. See
# https://gitlab.tiker.net/inducer/loopy/issues/85
VALID_NOSYNC_SCOPES = frozenset(["local", "global", "any"])


class SubscriptIndicesIsIntChecker(TypeReader):
    def map_subscript(self, expr):
        for idx in expr.index_tuple:
            type_inf_result = self.rec(idx)
            if not type_inf_result:
                raise LoopyError(
                        "When checking that subscript indices are integral: "
                        "Type inference did not find type of '%s'"
                        % idx)
            if not type_inf_result[0].is_integral():
                raise LoopyError("Non-integral array indices obtained in"
                        " {}.".format(expr))

        return self.rec(expr.aggregate)


def _check_for_integer_subscript_indices_inner(kernel, callables_table):

    from loopy.kernel.data import auto
    if any(arg.dtype in [None, auto] for arg in kernel.args) or (
            any(tv.dtype in [None, auto]
                for tv in kernel.temporary_variables.values())):
        # some types are not resolved => do not check.
        return

    from pymbolic.primitives import Subscript
    idx_int_checker = SubscriptIndicesIsIntChecker(kernel, callables_table)
    for insn in kernel.instructions:
        if isinstance(insn, MultiAssignmentBase):
            idx_int_checker(insn.expression, return_tuple=isinstance(insn,
                CallInstruction), return_dtype_set=True)
            [idx_int_checker(assignee) for assignee in insn.assignees if
                    isinstance(assignee, Subscript)]
        elif isinstance(insn, (CInstruction, _DataObliviousInstruction)):
            pass
        else:
            raise NotImplementedError("Unknown insn type %s." % (
                type(insn).__name__))


def check_for_integer_subscript_indices(t_unit):
    """
    Checks if every array access is of type :class:`int`.
    """
    from loopy.kernel.function_interface import CallableKernel, ScalarCallable
    for clbl in t_unit.callables_table.values():
        if isinstance(clbl, CallableKernel):
            _check_for_integer_subscript_indices_inner(clbl.subkernel,
                                                       t_unit.callables_table)
        elif isinstance(clbl, ScalarCallable):
            pass
        else:
            raise NotImplementedError(type(clbl).__name__)


@check_each_kernel
def check_sub_array_ref_inames_not_within_or_redn_inames(kernel: LoopKernel) -> None:
    all_within_inames = frozenset().union(*(insn.within_inames
                                            for insn in kernel.instructions))
    all_redn_inames = frozenset().union(*(insn.reduction_inames()
                                          for insn in kernel.instructions))
    all_sar_inames = frozenset().union(*(insn.sub_array_ref_inames()
                                         for insn in kernel.instructions))

    if all_sar_inames & all_within_inames:
        sample = next(iter(all_sar_inames & all_within_inames))
        raise LoopyError(f"Iname '{sample}' used as a sub-array ref's sweep"
                         " iname and an instruction's within inames. Such usage"
                         " is illegal.")

    if all_sar_inames & all_redn_inames:
        sample = next(iter(all_sar_inames & all_within_inames))
        raise LoopyError(f"Iname '{sample}' used as a sub-array ref's sweep"
                         " iname and a reduction iname. Such usage is"
                         " illegal.")


@check_each_kernel
def check_insn_attributes(kernel: LoopKernel) -> None:
    """
    Check for legality of attributes of every instruction in *kernel*.
    """
    all_insn_ids = {insn.id for insn in kernel.instructions}

    for insn in kernel.instructions:
        if not insn.within_inames <= kernel.all_inames():
            raise LoopyError("insn '%s' has unknown forced iname "
                    "dependencies: %s"
                    % (insn.id, ", ".join(
                        insn.within_inames - kernel.all_inames())))

        if insn.depends_on is not None and not insn.depends_on <= all_insn_ids:
            raise LoopyError("insn '%s' has unknown instruction "
                    "dependencies: %s"
                    % (insn.id, ", ".join(
                        insn.depends_on - all_insn_ids)))

        no_sync_with_insn_ids = {id for id, scope in insn.no_sync_with}
        if not no_sync_with_insn_ids <= all_insn_ids:
            raise LoopyError("insn '%s' has nosync directive with unknown "
                    "instruction ids: %s"
                    % (insn.id,
                       ", ".join(no_sync_with_insn_ids - all_insn_ids)))

        no_sync_with_scopes = {scope for id, scope in insn.no_sync_with}
        if not no_sync_with_scopes <= VALID_NOSYNC_SCOPES:
            raise LoopyError("insn '%s' has invalid nosync scopes: %s"
                    % (insn.id,
                       ", ".join(no_sync_with_scopes - VALID_NOSYNC_SCOPES)))


@check_each_kernel
def check_for_duplicate_insn_ids(knl: LoopKernel) -> None:
    """
    Check if multiple instructions of *knl* have the same
    :attr:`loopy.InstructionBase.id`.
    """
    insn_ids = set()

    for insn in knl.instructions:
        if not isinstance(insn.id, str):
            raise LoopyError("instruction id %r is not a string" % insn.id)
        if insn.id in insn_ids:
            raise LoopyError("duplicate instruction id: '%s'" % insn.id)
        insn_ids.add(insn.id)


@check_each_kernel
def check_loop_priority_inames_known(kernel: LoopKernel) -> None:
    """
    Checks if the inames in :attr:`loopy.LoopKernel.loop_priority` are part of
    the *kernel*'s domain.
    """
    for prio in kernel.loop_priority:
        for iname in prio:
            if iname not in kernel.all_inames():
                raise LoopyError("unknown iname '%s' in loop priorities" % iname)


@check_each_kernel
def check_multiple_tags_allowed(kernel: LoopKernel) -> None:
    """
    Checks if a multiple tags of an iname are compatible.
    """
    from loopy.kernel.data import (
        ForceSequentialTag,
        GroupInameTag,
        IlpBaseTag,
        LocalInameTag,
        UnrollTag,
        VectorizeTag,
        filter_iname_tags_by_type,
    )
    illegal_combinations = [
        (GroupInameTag, LocalInameTag, VectorizeTag, UnrollTag, ForceSequentialTag),
        (IlpBaseTag, ForceSequentialTag)
    ]
    for iname in kernel.inames.values():
        for comb in illegal_combinations:
            if len(filter_iname_tags_by_type(iname.tags, comb)) > 1:
                raise LoopyError("iname {} has illegal combination of "
                                 "tags: {}".format(iname.name, iname.tags))


def _check_for_double_use_of_hw_axes_inner(
            kernel: LoopKernel,
            callables_table: CallablesTable
        ) -> None:
    from loopy.kernel.data import GroupInameTag, LocalInameTag, UniqueInameTag
    from loopy.kernel.instruction import CallInstruction
    from loopy.symbolic import ResolvedFunction

    for insn in kernel.instructions:
        insn_tag_keys = set()
        if isinstance(insn, CallInstruction):
            assert isinstance(insn.expression.function, ResolvedFunction)
            clbl = callables_table[insn.expression.function.name]
            gsize, lsize = clbl.get_used_hw_axes(callables_table)
            insn_tag_keys |= {GroupInameTag(i).key for i in gsize}
            insn_tag_keys |= {LocalInameTag(i).key for i in lsize}

        for iname in insn.within_inames:
            for tag in kernel.iname_tags_of_type(iname, UniqueInameTag):
                key = tag.key
                if key in insn_tag_keys:
                    raise LoopyError("instruction '%s' has multiple "
                            "inames tagged '%s'" % (insn.id, tag))

                insn_tag_keys.add(key)


def check_for_double_use_of_hw_axes(t_unit: TranslationUnit) -> None:
    """
    Check if any instruction of *kernel* is within multiple inames tagged with
    the same hw axis tag.
    """
    from loopy.kernel.function_interface import CallableKernel, ScalarCallable
    for clbl in t_unit.callables_table.values():
        if isinstance(clbl, CallableKernel):
            _check_for_double_use_of_hw_axes_inner(clbl.subkernel,
                                                   t_unit.callables_table)
        elif isinstance(clbl, ScalarCallable):
            pass
        else:
            raise NotImplementedError(type(clbl).__name__)


@check_each_kernel
def check_for_inactive_iname_access(kernel: LoopKernel) -> None:
    """
    Check if any instruction accesses an iname but is not within it.
    """
    for insn in kernel.instructions:
        expression_inames = insn.read_dependency_names() & kernel.all_inames()

        if not expression_inames <= insn.within_inames:
            raise LoopyError(
                    "instruction '%s' references "
                    "inames '%s' that the instruction does not depend on in "
                    "the kernel '%s'"
                    % (insn.id,
                        ", ".join(expression_inames
                                  - insn.within_inames), kernel.name))


@check_each_kernel
def check_for_unused_inames(kernel: LoopKernel) -> None:
    """
    Check if there are any unused inames in the kernel.
    """
    # Warn if kernel has unused inames
    from loopy.transform.iname import get_used_inames
    unused_inames = kernel.all_inames() - get_used_inames(kernel)
    if unused_inames:
        warn_with_kernel(
            kernel, "unused_inames",
            "Found unused inames in kernel: %s "
            "Unused inames during linearization will be prohibited in "
            "Loopy version 2021.X."
            % unused_inames)


def _is_racing_iname_tag(tv: TemporaryVariable, tag: InameImplementationTag) -> bool:
    from loopy.kernel.data import (
        AddressSpace,
        ConcurrentTag,
        GroupInameTag,
        LocalInameTagBase,
        auto,
    )

    if tv.address_space == AddressSpace.PRIVATE:
        return (
                isinstance(tag, ConcurrentTag)
                and not isinstance(tag, (LocalInameTagBase, GroupInameTag)))

    elif tv.address_space == AddressSpace.LOCAL:
        return (
                isinstance(tag, ConcurrentTag)
                and not isinstance(tag, GroupInameTag))

    elif tv.address_space == AddressSpace.GLOBAL:
        return isinstance(tag, ConcurrentTag)

    elif tv.address_space == auto:
        raise LoopyError("scope of temp var '%s' has not yet been"
                "determined" % tv.name)

    else:
        raise ValueError("unexpected value of temp_var.address_space for "
                "temporary variable '%s'" % tv.name)


@check_each_kernel
def check_for_write_races(kernel: LoopKernel) -> None:
    """
    Check if any memory accesses lead to write races.
    """
    from loopy.kernel.data import ConcurrentTag

    for insn in kernel.instructions:
        for assignee_name, assignee_indices in zip(
                insn.assignee_var_names(),
                insn.assignee_subscript_deps()):
            assignee_inames = assignee_indices & kernel.all_inames()
            if not assignee_inames <= insn.within_inames:
                raise LoopyError(
                        "assignee of instructions '%s' references "
                        "iname that the instruction does not depend on"
                        % insn.id)

            if assignee_name in kernel.arg_dict:
                # Any concurrent tags that are not depended upon by the assignee
                # will cause write races.

                raceable_parallel_insn_inames = {
                    iname for iname in insn.within_inames
                    if kernel.iname_tags_of_type(iname, ConcurrentTag)}

            elif assignee_name in kernel.temporary_variables:
                temp_var = kernel.temporary_variables[assignee_name]
                raceable_parallel_insn_inames = {
                        iname for iname in insn.within_inames
                        if any(_is_racing_iname_tag(temp_var, tag)
                            for tag in kernel.iname_tags(iname))}

            else:
                raise LoopyError("invalid assignee name in instruction '%s'"
                        % insn.id)

            race_inames = \
                    raceable_parallel_insn_inames - assignee_inames

            if race_inames:
                warn_with_kernel(kernel, "write_race(%s)" % insn.id,
                        "instruction '%s' contains a write race: "
                        "instruction will be run across parallel iname(s) "
                        "'%s', which is/are not referenced in the lhs index"
                        % (insn.id, ",".join(race_inames)),
                        WriteRaceConditionWarning)


@check_each_kernel
def check_for_data_dependent_parallel_bounds(kernel: LoopKernel) -> None:
    """
    Check that inames tagged as hw axes have bounds that are known at kernel
    launch.
    """
    from loopy.kernel.data import ConcurrentTag

    for i, dom in enumerate(kernel.domains):
        dom_inames = set(dom.get_var_names(dim_type.set))
        par_inames = {
                iname for iname in dom_inames
                if kernel.iname_tags_of_type(iname, ConcurrentTag)}

        if not par_inames:
            continue

        parameters = set(dom.get_var_names(dim_type.param))
        for par in parameters:
            if par in kernel.temporary_variables:
                raise LoopyError("Domain number %d has a data-dependent "
                        "parameter '%s' and contains parallel "
                        "inames '%s'. This is not allowed (for now)."
                        % (i, par, ", ".join(par_inames)))


# {{{ check access bounds

# {{{ helpers for _AccessCheckMapper

def _align_and_intersect(d1, d2):
    d1, d2 = isl.align_two(d1, d2)
    return (d1 & d2).params()


def _align_and_intersect_with_caller_assumption(callee_assumptions,
                                                caller_assumptions):

    for name, (dt, pos) in caller_assumptions.get_var_dict().items():
        caller_assumptions = caller_assumptions.set_dim_name(
            dt, pos, f"_lp_caller_{name}")

    return _align_and_intersect(callee_assumptions,
                                caller_assumptions)


def _mark_variables_from_caller(expr):
    import pymbolic.primitives as prim

    from loopy.symbolic import SubstitutionMapper

    def subst_func(x):
        if isinstance(x, prim.Variable):
            return prim.Variable(f"_lp_caller_{x.name}")

    return SubstitutionMapper(subst_func)(expr)

# }}}


class _AccessCheckMapper(WalkMapper):
    def __init__(self, kernel, callables_table):
        self.kernel = kernel
        self.callables_table = callables_table
        super().__init__()

    @memoize_method
    def _make_slab(self, space, iname, start, stop):
        from loopy.isl_helpers import make_slab
        return make_slab(space, iname, start, stop)

    @memoize_method
    def _get_access_range(self, domain, subscript):
        from loopy.symbolic import UnableToDetermineAccessRangeError, get_access_map
        try:
            return get_access_map(domain, subscript).range()
        except UnableToDetermineAccessRangeError:
            return None

    def map_subscript(self, expr, domain, insn_id):
        WalkMapper.map_subscript(self, expr, domain, insn_id)

        from pymbolic.primitives import Variable
        assert isinstance(expr.aggregate, Variable)

        shape = None
        var_name = expr.aggregate.name
        if var_name in self.kernel.arg_dict:
            arg = self.kernel.arg_dict[var_name]
            shape = arg.shape
        elif var_name in self.kernel.temporary_variables:
            tv = self.kernel.temporary_variables[var_name]
            shape = tv.shape

        if shape is not None:
            subscript = expr.index

            if not isinstance(subscript, tuple):
                subscript = (subscript,)

            from loopy.symbolic import get_dependencies

            available_vars = set(domain.get_var_dict())
            shape_deps = set()
            for shape_axis in shape:
                if shape_axis is not None:
                    shape_deps.update(get_dependencies(shape_axis))

            if not (get_dependencies(subscript) <= available_vars
                    and shape_deps <= available_vars):
                return

            if len(subscript) != len(shape):
                raise LoopyError("subscript to '%s' in '%s' has the wrong "
                        "number of indices (got: %d, expected: %d)" % (
                            expr.aggregate.name, expr,
                            len(subscript), len(shape)))

            access_range = self._get_access_range(domain, subscript)
            if access_range is None:
                # Likely: index was non-affine, nothing we can do.
                return

            shape_domain = isl.BasicSet.universe(access_range.get_space())
            for idim in range(len(subscript)):
                shape_axis = shape[idim]

                if shape_axis is not None:
                    slab = self._make_slab(
                            shape_domain.get_space(), (dim_type.in_, idim),
                            0, shape_axis)

                    shape_domain = shape_domain.intersect(slab)

            if not access_range.is_subset(shape_domain):
                raise LoopyIndexError("'%s' in instruction '%s' "
                        "accesses out-of-bounds array element (could not"
                        " establish '%s' is a subset of '%s')."
                        % (expr, insn_id, access_range, shape_domain))

    def map_if(self, expr, domain, insn_id):
        from loopy.symbolic import condition_to_set
        then_set = condition_to_set(domain.space, expr.condition)
        if then_set is None:
            # condition cannot be inferred as ISL expression => ignore
            # for domain contributions enforced by it
            then_set = else_set = isl.BasicSet.universe(domain.space)
        else:
            else_set = then_set.complement()

        self.rec(expr.then, domain & then_set, insn_id)
        self.rec(expr.else_, domain & else_set, insn_id)

    def map_call(self, expr, domain, insn_id):
        # perform access checks on the call arguments
        super().map_call(expr, domain, insn_id)

        import pymbolic.primitives as prim

        from loopy.diagnostic import ExpressionToAffineConversionError
        from loopy.kernel.function_interface import (
            CallableKernel,
            get_kw_pos_association,
        )
        from loopy.symbolic import get_dependencies, guarded_aff_from_expr

        if (isinstance(expr.function, ResolvedFunction)
            and isinstance(self.callables_table[expr.function.name],
                           CallableKernel)):

            subkernel = self.callables_table[expr.function.name].subkernel

            # The plan here is to add the constraints coming from the values
            # args passed at a call-site as assumptions to the callee. To avoid
            # variable-naming collision between the caller and callee, we
            # prepend all the caller variable by "_lp_caller_".

            # {{{ get bset from kwargs

            kw_to_pos, _ = get_kw_pos_association(subkernel)

            arg_id_to_arg = self.kernel.id_to_insn[insn_id].arg_id_to_arg()

            kwargs = {k: _mark_variables_from_caller(arg_id_to_arg[kw_to_pos[k]])
                      for k in subkernel.get_unwritten_value_args()}

            kw_space = isl.Space.create_from_names(
                subkernel.isl_context, set=[],
                params=(get_dependencies(tuple(kwargs.values()))
                        | set(kwargs.keys())))

            extra_assumptions = isl.BasicSet.universe(kw_space).params()

            for kw, arg in kwargs.items():
                try:
                    aff = guarded_aff_from_expr(extra_assumptions.space,
                                                prim.Variable(kw) - arg)
                except ExpressionToAffineConversionError:
                    # arg expression not affine => don't add any constraints
                    # corresponding to it
                    continue

                extra_assumptions = (extra_assumptions
                                     .add_constraint(isl.Constraint
                                                     .equality_from_aff(aff)))

            # FIXME: caller inames could be arguments => should take that into
            # account as well

            # take into account caller's assumptions
            extra_assumptions = _align_and_intersect_with_caller_assumption(
                extra_assumptions, self.kernel.assumptions)

            # project out the assumptions on caller's variables as they don't
            # bear any semantic meaning in the callee.
            extra_assumptions = extra_assumptions.project_out_except(
                types=[isl.dim_type.param],
                names=subkernel.get_unwritten_value_args())

            subkernel = subkernel.copy(
                assumptions=_align_and_intersect(subkernel.assumptions,
                                                 extra_assumptions))

            # }}}

            _check_bounds_inner_rec(subkernel, self.callables_table)


def _check_bounds_inner(kernel: LoopKernel, callables_table: CallablesTable) -> None:
    from loopy.kernel.instruction import get_insn_domain

    temp_var_names = set(kernel.temporary_variables)
    acm = _AccessCheckMapper(kernel, callables_table)
    kernel_assumptions_is_universe = kernel.assumptions.is_universe()
    for insn in kernel.instructions:
        domain = get_insn_domain(insn, kernel)

        # data-dependent bounds? can't do much
        if set(domain.get_var_names(dim_type.param)) & temp_var_names:
            continue

        if kernel_assumptions_is_universe:
            domain_with_assumptions = domain
        else:
            domain, assumptions = isl.align_two(domain, kernel.assumptions)
            domain_with_assumptions = domain & assumptions

        def run_acm(expr):
            acm(expr, domain_with_assumptions, insn.id)  # noqa: B023
            return expr

        insn.with_transformed_expressions(run_acm)


def _check_bounds_inner_rec(
            kernel: LoopKernel,
            callables_table: CallablesTable
        ) -> None:
    if kernel.options.enforce_array_accesses_within_bounds not in [
            "no_check",
            True,
            False]:
        raise LoopyError("invalid value for option "
                "'enforce_array_accesses_within_bounds': %s"
                % kernel.options.enforce_array_accesses_within_bounds)

    if kernel.options.enforce_array_accesses_within_bounds == "no_check":
        return

    from pytools import ProcessLogger
    with ProcessLogger(logger, "%s: check array access within bounds" % kernel.name):
        if kernel.options.enforce_array_accesses_within_bounds:
            _check_bounds_inner(kernel, callables_table)
        else:
            from loopy.diagnostic import LoopyIndexError
            try:
                _check_bounds_inner(kernel, callables_table)
            except LoopyIndexError as e:
                from loopy.diagnostic import warn_with_kernel
                warn_with_kernel(kernel, "array_access_out_of_bounds", str(e))


def check_bounds(t_unit: TranslationUnit) -> None:
    """
    Performs out-of-bound check for every array access.
    """
    for epoint in t_unit.entrypoints:
        _check_bounds_inner_rec(t_unit[epoint],
                                t_unit.callables_table)

# }}}


# {{{ check write destinations

@check_each_kernel
def check_write_destinations(kernel: LoopKernel) -> None:
    for insn in kernel.instructions:
        for wvar in insn.assignee_var_names():
            if wvar in kernel.all_inames():
                raise LoopyError("iname '%s' may not be written" % wvar)

            insn_domain = kernel.get_inames_domain(insn.within_inames)
            insn_params = set(insn_domain.get_var_names(dim_type.param))

            if wvar in kernel.all_params():
                if wvar not in kernel.temporary_variables:
                    raise LoopyError("domain parameter '%s' may not be written"
                            "--it is not a temporary variable" % wvar)

                if wvar in insn_params:
                    raise LoopyError("domain parameter '%s' may not be written "
                            "inside a domain dependent on it" % wvar)

            if not (wvar in kernel.temporary_variables
                    or wvar in kernel.arg_dict) and wvar not in kernel.all_params():
                raise LoopyError

# }}}


# {{{ check_has_schedulable_iname_nesting

@check_each_kernel
def check_has_schedulable_iname_nesting(kernel: LoopKernel) -> None:
    from loopy.transform.iname import (
        get_iname_duplication_options,
        has_schedulable_iname_nesting,
    )
    if not has_schedulable_iname_nesting(kernel):
        import itertools as it
        opt = get_iname_duplication_options(kernel)
        opt_str = "\n".join(f"* Duplicate {i} within instructions {w}"
                            for i, w in it.islice(opt, 3))
        raise LoopyError("Kernel does not have a schedulable iname nesting. "
                "In order for there to exist a feasible loop nesting, you "
                "may need to duplicate an iname. To do so, call "
                "loopy.duplicate_iname. Use loopy.get_iname_duplication_options "
                "to get hints about which iname to duplicate. Here are some "
                "options:\n%s" % opt_str)

# }}}


# {{{ check_variable_access_ordered

def declares_nosync_with(kernel, var_address_space, dep_a, dep_b):
    dep_a = kernel.id_to_insn[dep_a]
    dep_b = kernel.id_to_insn[dep_b]
    from loopy.kernel.data import AddressSpace
    if var_address_space == AddressSpace.GLOBAL:
        search_scopes = ["global", "any"]
    elif var_address_space == AddressSpace.LOCAL:
        search_scopes = ["local", "any"]
    elif var_address_space == AddressSpace.PRIVATE:
        search_scopes = ["any"]
    else:
        raise ValueError("unexpected value of 'AddressSpace'")

    ab_nosync = False
    ba_nosync = False

    for scope in search_scopes:
        if (dep_a.id, scope) in dep_b.no_sync_with:
            ab_nosync = True
        if (dep_b.id, scope) in dep_a.no_sync_with:
            ba_nosync = True

    return ab_nosync and ba_nosync


def _get_address_space(kernel: LoopKernel, var: str) -> AddressSpace | type[auto]:
    from loopy.kernel.data import ArrayArg, ValueArg
    if var in kernel.temporary_variables:
        address_space = kernel.temporary_variables[var].address_space
    else:
        arg = kernel.arg_dict[var]
        if isinstance(arg, ArrayArg):
            address_space = arg.address_space
        elif isinstance(arg, ValueArg):
            address_space = AddressSpace.PRIVATE
        else:
            # No need to consider ConstantArg and ImageArg (for now)
            # because those won't be written.
            raise ValueError("could not determine address_space of '%s'" % var)
    return address_space


def _get_topological_order(kernel: LoopKernel) -> Sequence[str]:
    """
    Returns a :class:`list` of insn ids of *kernel* in a topological sort
    order.

    If there is a dependency cycle within the instructions of *kernel* raises a
    :class:`loopy.diagnostic.DependencyCycleFound` exception.
    """
    from pytools.graph import compute_sccs

    from loopy.diagnostic import DependencyCycleFound

    dep_map = {insn.id: insn.depends_on for insn in kernel.instructions}

    # pytools.graph.compute_sccs serves 2 purposes:
    #   1. computes topological sort order of instructions.
    #   2. provides info. about any cycles in the graph.
    sccs = compute_sccs(dep_map)
    order = []

    for scc in sccs:
        if len(scc) != 1:
            raise DependencyCycleFound(", ".join(scc))
        order.append(scc[0])

    return order


def _check_variable_access_ordered_inner(kernel: LoopKernel) -> None:
    from loopy.kernel.tools import find_aliasing_equivalence_classes
    from loopy.symbolic import AccessRangeOverlapChecker
    overlap_checker = AccessRangeOverlapChecker(kernel)
    aliasing_equiv_classes = find_aliasing_equivalence_classes(kernel)

    # dep_reqs_to_vars: A mapping (writer_id, dep_req_id) -> set of variable names,
    # where the tuple denotes a pair of instructions IDs, and the variable
    # names are the ones that necessitate a dependency.
    #
    # This mapping describes all the pairs of instructions (involving at least
    # one write) that require ordering by way of a dependency.
    # This mapping is then "whittled down" to remove pairs that have
    # dependencies between them. E.g. a pair of writers will be contained in
    # the mapping in both directions.
    #
    # Note: This can be worst-case O(n^2) in the number of instructions.
    dep_reqs_to_vars: dict[tuple[str, str], set[str]] = {}

    wmap = kernel.writer_map()
    rmap = kernel.reader_map()

    # {{{ populate 'dep_reqs_to_vars'

    for var in kernel.get_written_variables():
        address_space = _get_address_space(kernel, var)
        eq_class = aliasing_equiv_classes[var]

        readers = set.union(
                *[rmap.get(eq_name, set()) for eq_name in eq_class])
        writers = set.union(
                *[wmap.get(eq_name, set()) for eq_name in eq_class])

        for writer in writers:
            required_deps = (readers | writers) - {writer}
            required_deps = {req_dep
                for req_dep in required_deps
                if not declares_nosync_with(kernel, address_space, writer,
                    req_dep)}

            for req_dep in required_deps:
                dep_reqs_to_vars.setdefault((writer, req_dep), set()).add(var)

    # }}}

    # {{{ compute rev_depends, depends_on

    # depends_on: mapping from insn_ids to their dependencies
    depends_on: dict[str, set[str]] = {
        not_none(insn.id): set() for insn in kernel.instructions}
    # rev_depends: mapping from insn_ids to their reverse deps.
    rev_depends: dict[str, set[str]] = {
        not_none(insn.id): set() for insn in kernel.instructions}

    for insn in kernel.instructions:
        depends_on[not_none(insn.id)].update(insn.depends_on)
        for dep in insn.depends_on:
            rev_depends[dep].add(not_none(insn.id))

    # }}}

    # {{{ remove pairs from dep_reqs_to_vars for which dependencies exist

    topological_order = _get_topological_order(kernel)

    # TODO: Type this
    def satisfy_dep_reqs_in_order(dep_reqs_to_vars, edges, order) -> None:
        """
        Considering a graph defined by *edges* (as ``key -> value``),
        remove pairs of nodes from *dep_reqs_to_vars* for which edges
        exist in the graph. In doing so, make use of the topological
        order *order* (given as a sequence of *nodes*). For ``i<j``,
        no direct or indirect edges from ``order[j]`` to ``order[i]``
        exist. (All edges go from 'low index' to 'high index'.)
        """

        insn_to_req_deps = defaultdict(set)
        insn_to_order = {insn: i for i, insn in enumerate(order)}

        # Use dep_reqs_to_vars to construct insn_to_req_deps, listing
        # path endpoints pred -> ... -> insn to be considered for elimination.
        for insn, pred in dep_reqs_to_vars:
            if insn_to_order[pred] > insn_to_order[insn]:
                # If *pred* happens after *insn*, then there is no path
                # *pred* -> ... -> *insn*.
                continue

            insn_to_req_deps[pred].add(insn)

        for pred, check_successors in insn_to_req_deps.items():
            # for each *pred*, we will calculate all the direct/indirect
            # instructions that can be reached.
            seen_successors = set()
            # first let us start with direct successors
            to_check = edges[pred].copy()
            while to_check:
                successor = to_check.pop()
                seen_successors.add(successor)

                # Next we check if this successor was in *dep_reqs_to_vars*
                # and remove the pair from there if it is
                if successor in check_successors:
                    dep_reqs_to_vars.pop((successor, pred))
                    check_successors.remove(successor)
                    if not check_successors:
                        break

                # next we iterate through the successors of successor.
                for edge in edges[successor]:
                    if edge not in seen_successors:
                        to_check.add(edge)

    # forward dep. graph traversal in reverse topological sort order
    # (proceeds "end of program" -> "beginning of program")
    satisfy_dep_reqs_in_order(dep_reqs_to_vars, depends_on,
            topological_order[::-1])

    # Run the same thing on the reversed graph, with the reverse topological_order.
    # All edges above are 'low index to high index'.
    # Reversing graph and order maintains this property.

    # reverse dep. graph traversal in topological sort order
    # (proceeds "beginning of program" -> "end of program")
    satisfy_dep_reqs_in_order(dep_reqs_to_vars, rev_depends, topological_order)

    # }}}

    # {{{ handle dependency requirements that weren't satisfied

    for (writer_id, other_id), variables in dep_reqs_to_vars.items():
        writer = kernel.id_to_insn[writer_id]
        other = kernel.id_to_insn[other_id]

        for var in variables:
            eq_class = aliasing_equiv_classes[var]
            unaliased_readers = rmap.get(var, set())
            unaliased_writers = wmap.get(var, set())

            is_relationship_by_aliasing = not (
                writer_id in unaliased_writers
                and (writer_id in unaliased_writers
                    or other_id in unaliased_readers))

            # Do not enforce ordering for disjoint access ranges
            if (not is_relationship_by_aliasing and not
                overlap_checker.do_access_ranges_overlap_conservative(
                        writer_id, "w", other_id, "any", var)):
                continue

            # Do not enforce ordering for aliasing-based relationships
            # in different groups.
            if (is_relationship_by_aliasing and (
                    bool(writer.groups & other.conflicts_with_groups)
                    or
                    bool(other.groups & writer.conflicts_with_groups))):
                continue

            msg = ("No dependency relationship found between "
                    "'{writer_id}' which writes {var} and "
                    "'{other_id}' which also accesses {var}. "
                    "Either add a (possibly indirect) dependency "
                    "between the two, or add them to each others' nosync "
                    "set to indicate that no ordering is intended, or "
                    "turn off this check by setting the "
                    "'enforce_variable_access_ordered' option "
                    "(more issues of this type may exist--only reporting "
                    "the first one)"
                    .format(
                        writer_id=writer_id,
                        other_id=other_id,
                        var=(
                            "the variable '%s'" % var
                            if len(eq_class) == 1
                            else (
                                "the aliasing equivalence class '%s'"
                                % ", ".join(eq_class))
                            )))

            from loopy.diagnostic import VariableAccessNotOrdered
            raise VariableAccessNotOrdered(msg)

    # }}}


@check_each_kernel
def check_variable_access_ordered(kernel: LoopKernel) -> None:
    """Checks that between each write to a variable and all other accesses to
    the variable there is either:

    * a direct/indirect dependency edge, or
    * an explicit statement that no ordering is necessary (expressed
      through a bi-directional :attr:`loopy.InstructionBase.no_sync_with`)
    """

    if kernel.options.enforce_variable_access_ordered not in [
            "no_check",
            True,
            False]:
        raise LoopyError("invalid value for option "
                "'enforce_variable_access_ordered': %s"
                % kernel.options.enforce_variable_access_ordered)

    if kernel.options.enforce_variable_access_ordered == "no_check":
        return

    from pytools import ProcessLogger
    with ProcessLogger(logger, "%s: check variable access ordered" % kernel.name):
        if kernel.options.enforce_variable_access_ordered:
            _check_variable_access_ordered_inner(kernel)
        else:
            from loopy.diagnostic import VariableAccessNotOrdered
            try:
                _check_variable_access_ordered_inner(kernel)
            except VariableAccessNotOrdered as e:
                from loopy.diagnostic import warn_with_kernel
                warn_with_kernel(kernel, "variable_access_ordered", str(e))

# }}}

# }}}


def pre_schedule_checks(t_unit: TranslationUnit) -> None:
    try:
        logger.debug("pre-schedule checks start for entrypoints: "
                     f"{t_unit.entrypoints}.")

        check_for_integer_subscript_indices(t_unit)

        check_functions_are_resolved(t_unit)
        check_separated_array_consistency(t_unit)
        check_offsets_and_dim_tags(t_unit)
        # Ordering restriction:
        # check_sub_array_ref_inames_not_within_or_redn_inames should be done
        # before check_bounds. See: BatchedAccessMapMapper.map_sub_array_ref.
        check_sub_array_ref_inames_not_within_or_redn_inames(t_unit)
        check_for_duplicate_insn_ids(t_unit)
        check_for_double_use_of_hw_axes(t_unit)
        check_insn_attributes(t_unit)
        check_loop_priority_inames_known(t_unit)
        check_multiple_tags_allowed(t_unit)
        check_for_inactive_iname_access(t_unit)
        check_for_unused_inames(t_unit)
        check_for_write_races(t_unit)
        check_for_data_dependent_parallel_bounds(t_unit)
        check_bounds(t_unit)
        check_write_destinations(t_unit)
        check_has_schedulable_iname_nesting(t_unit)
        check_variable_access_ordered(t_unit)

        logger.debug("pre-schedule checks done")
    except KeyboardInterrupt:
        raise
    except Exception:
        print(75*"=")
        print("failing translation unit during pre-schedule check:")
        print(75*"=")
        print(t_unit)
        print(75*"=")
        raise


# {{{ post-schedule / pre-code-generation checks

# {{{ check_for_nested_base_storage

def check_for_nested_base_storage(kernel: LoopKernel) -> None:
    # must run after preprocessing has created variables for base_storage

    from loopy.kernel.data import ArrayArg
    arrays: list[ArrayBase] = [
        arg for arg in kernel.args if isinstance(arg, ArrayArg)
        ]
    arrays = arrays + list(kernel.temporary_variables.values())

    name_to_array = {ary.name: ary for ary in arrays}

    for ary in kernel.temporary_variables.values():
        if ary.base_storage:
            storage_array = name_to_array.get(ary.base_storage, None)

            if storage_array is None:
                raise LoopyError("Nothing known about storage array "
                        f"'{ary.base_storage}' serving as base_storage of "
                        f"'{ary.name}'. "
                        "(Note: base storage is no longer automatically allocated. "
                        "Call allocate_temporaries_for_base_storage to automatically "
                        "allocate.)")

            if storage_array.base_storage:
                raise ValueError("storage array "
                        f"'{storage_array.name}' serving as base_storage of "
                        f"'{ary.name}' may not itself use base_storage "
                        "(currently given as '{storage_array.base_storage}'")

# }}}


# {{{ check for unused hw axes

def _check_for_unused_hw_axes_in_kernel_chunk(
            kernel: LoopKernel,
            callables_table: CallablesTable,
            sched_index: int | None = None
        ) -> int:
    from loopy.schedule import (
        Barrier,
        CallKernel,
        EnterLoop,
        LeaveLoop,
        ReturnFromKernel,
        RunInstruction,
        gather_schedule_block,
        get_insn_ids_for_block_at,
    )
    assert kernel.linearization is not None

    if sched_index is None:
        group_axes = set()
        local_axes = set()

        i = 0
        loop_end_i = past_end_i = len(kernel.linearization)
    else:
        assert isinstance(kernel.linearization[sched_index], CallKernel)
        _, past_end_i = gather_schedule_block(kernel.linearization, sched_index)
        group_size, local_size = kernel.get_grid_sizes_for_insn_ids_as_exprs(
                get_insn_ids_for_block_at(kernel.linearization, sched_index),
                callables_table, return_dict=True)

        group_axes = set(group_size.keys())
        local_axes = set(local_size.keys())

        i = sched_index + 1
        assert isinstance(kernel.linearization[past_end_i - 1], ReturnFromKernel)
        loop_end_i = past_end_i - 1

    # alternative: just disregard length-1 dimensions?

    from loopy.kernel.data import AutoLocalInameTagBase, GroupInameTag, LocalInameTag

    while i < loop_end_i:
        sched_item = kernel.linearization[i]
        if isinstance(sched_item, CallKernel):
            i = _check_for_unused_hw_axes_in_kernel_chunk(kernel,
                    callables_table, i)

        elif isinstance(sched_item, RunInstruction):
            insn = kernel.id_to_insn[sched_item.insn_id]
            i += 1

            if isinstance(insn, NoOpInstruction):
                continue

            group_axes_used = set()
            local_axes_used = set()

            for iname in insn.within_inames:
                ltags = kernel.iname_tags_of_type(iname, LocalInameTag, max_num=1)
                gtags = kernel.iname_tags_of_type(iname, GroupInameTag, max_num=1)
                altags = kernel.iname_tags_of_type(
                        iname, AutoLocalInameTagBase, max_num=1)

                if ltags:
                    tag: AxisTag
                    tag, = ltags
                    local_axes_used.add(tag.axis)
                elif gtags:
                    tag, = gtags
                    group_axes_used.add(tag.axis)
                elif altags:
                    raise LoopyError("auto local tag encountered")

            # {{{ account for any hw axes due to a callable

            if isinstance(insn, CallInstruction):
                assert isinstance(insn.expression.function, ResolvedFunction)
                clbl = callables_table[insn.expression.function.name]
                clbl_g_axes, clbl_l_axes = clbl.get_used_hw_axes(callables_table)
                assert len(group_axes_used & clbl_g_axes) == 0
                assert len(local_axes_used & clbl_l_axes) == 0
                group_axes_used |= clbl_g_axes
                local_axes_used |= clbl_l_axes

            # }}}

            if group_axes != group_axes_used:
                raise LoopyError(
                        f"instruction '{insn.id}' does not use all group hw axes "
                        "(available: %s used: %s). "
                        "Calling loopy.add_inames_for_unused_hw_axes(...) "
                        "might help."
                        % (
                            ",".join(str(i) for i in group_axes),
                            ",".join(str(i) for i in group_axes_used)))

            if local_axes != local_axes_used:
                raise LoopyError(
                        f"instruction '{insn.id}' does not use all local hw axes "
                        "(available: %s used: %s). "
                        "Calling loopy.add_inames_for_unused_hw_axes(...) "
                        "might help."
                        % (
                            ",".join(str(i) for i in local_axes),
                            ",".join(str(i) for i in local_axes_used)))

        elif isinstance(sched_item, (Barrier, EnterLoop, LeaveLoop)):
            i += 1
            continue

        else:
            raise TypeError(
                    "schedule item not understood: %s" % type(sched_item).__name__)

    return past_end_i


def check_for_unused_hw_axes_in_insns(
            kernel: LoopKernel,
            callables_table: CallablesTable
        ) -> None:
    if kernel.linearization:
        _check_for_unused_hw_axes_in_kernel_chunk(kernel,
                callables_table)

# }}}


# {{{ check that atomic ops are used exactly on atomic arrays

def check_that_atomic_ops_are_used_exactly_on_atomic_arrays(
            kernel: LoopKernel
        ) -> None:
    from loopy.kernel.data import ArrayBase, Assignment
    from loopy.types import AtomicType
    atomicity_candidates = (
            {v.name for v in kernel.temporary_variables.values()
                if isinstance(v.dtype, AtomicType)}
            |
            {v.name for v in kernel.args
                if isinstance(v, ArrayBase)
                and isinstance(v.dtype, AtomicType)})

    for insn in kernel.instructions:
        if not isinstance(insn, Assignment):
            continue

        atomic_accesses = {a.var_name for a in insn.atomicity}
        if not atomic_accesses <= atomicity_candidates:
            raise LoopyError("atomic access in instruction '%s' to "
                    "non-atomic variable(s) '%s'"
                    % (insn.id,
                        ",".join(atomic_accesses - atomicity_candidates)))

        accessed_atomic_vars = insn.dependency_names() & atomicity_candidates
        if not accessed_atomic_vars <= atomic_accesses:
            raise LoopyError("atomic variable(s) '%s' in instruction '%s' "
                    "used in non-atomic access"
                    % (
                        ",".join(accessed_atomic_vars - atomic_accesses),
                        insn.id))

# }}}


# {{{ check that temporaries are defined in subkernels where used

def check_that_temporaries_are_defined_in_subkernels_where_used(
            kernel: LoopKernel
        ) -> None:
    from loopy.kernel.data import AddressSpace
    from loopy.kernel.tools import get_subkernels

    globally_defined_base_storage = set()

    for subkernel in get_subkernels(kernel):
        locally_defined_base_storage = set()

        from loopy.schedule.tools import (
            temporaries_read_in_subkernel,
            temporaries_written_in_subkernel,
        )

        for temporary in temporaries_written_in_subkernel(kernel, subkernel):
            tval = kernel.temporary_variables[temporary]
            if tval.base_storage is not None:
                if tval.address_space in (AddressSpace.PRIVATE, AddressSpace.LOCAL):
                    locally_defined_base_storage.add(tval.base_storage)
                else:
                    assert tval.address_space == AddressSpace.GLOBAL
                    globally_defined_base_storage.add(tval.base_storage)

        for temporary in (
                temporaries_read_in_subkernel(kernel, subkernel) -
                temporaries_written_in_subkernel(kernel, subkernel)):
            tval = kernel.temporary_variables[temporary]

            if tval.initializer is not None:
                continue

            # For aliased temporaries, check if there is an aliased definition.
            if tval.base_storage is not None:
                if tval.base_storage not in (locally_defined_base_storage
                                             | globally_defined_base_storage):
                    from loopy.diagnostic import MissingDefinitionError
                    raise MissingDefinitionError("temporary variable '%s' gets "
                            "used in subkernel '%s' and neither it nor its "
                            "aliases have a definition" % (temporary, subkernel))
                continue

            if tval.address_space in (AddressSpace.PRIVATE, AddressSpace.LOCAL):
                from loopy.diagnostic import MissingDefinitionError
                raise MissingDefinitionError("temporary variable '%s' gets used "
                        "in subkernel '%s' without a definition (maybe you forgot "
                        "to call loopy.save_and_reload_temporaries?)"
                        % (temporary, subkernel))

# }}}


# {{{ check that all instructions are scheduled

def check_that_all_insns_are_scheduled(kernel: LoopKernel) -> None:
    assert kernel.linearization is not None

    all_schedulable_insns = {not_none(insn.id) for insn in kernel.instructions}
    from loopy.schedule import sched_item_to_insn_id
    scheduled_insns = {
        insn_id
        for sched_item in kernel.linearization
        for insn_id in sched_item_to_insn_id(sched_item)}

    assert scheduled_insns <= all_schedulable_insns

    if scheduled_insns < all_schedulable_insns:
        from loopy.diagnostic import UnscheduledInstructionError
        raise UnscheduledInstructionError(
            "unscheduled instructions: '%s'"
            % ", ".join(all_schedulable_insns - scheduled_insns))

# }}}


# {{{ check that shapes and strides are arguments

def check_that_shapes_and_strides_are_arguments(kernel: LoopKernel) -> None:
    import loopy as lp
    from loopy.kernel.array import ArrayBase, FixedStrideArrayDimTag
    from loopy.kernel.data import ValueArg
    from loopy.symbolic import get_dependencies

    integer_arg_names = {
            arg.name
            for arg in kernel.args
            if isinstance(arg, ValueArg)
            and not_none(arg.dtype).is_integral()}

    for arg in kernel.args:
        if isinstance(arg, ArrayBase):
            if isinstance(arg.shape, tuple):
                shape_deps: set[str] = set()
                for shape_axis in arg.shape:
                    if shape_axis is not None:
                        shape_deps.update(get_dependencies(shape_axis))

                if not shape_deps <= integer_arg_names:
                    raise LoopyError("'%s' has a shape that depends on "
                            "non-argument(s): %s" % (
                                arg.name, ", ".join(shape_deps-integer_arg_names)))

            if arg.dim_tags is None:
                continue

            for dim_tag in arg.dim_tags:
                if isinstance(dim_tag, FixedStrideArrayDimTag):
                    if dim_tag.stride is lp.auto:
                        continue

                    deps = get_dependencies(dim_tag.stride)
                    if not deps <= integer_arg_names:
                        raise LoopyError("'%s' has a stride that depends on "
                                "non-argument(s): %s" % (
                                    arg.name, ", ".join(deps-integer_arg_names)))

# }}}


# {{{ validate_kernel_call_sites

def _get_sub_array_ref_swept_range(
            kernel: LoopKernel,
            sar: SubArrayRef
        ) -> isl.Set:
    from loopy.symbolic import get_access_map
    domain = kernel.get_inames_domain(frozenset({iname_var.name
                                                 for iname_var in sar.swept_inames}))
    return get_access_map(domain, sar.swept_inames, kernel.assumptions).range()


def _are_sub_array_refs_equivalent(
            sar1: SubArrayRef,
            sar2: SubArrayRef,
            caller: LoopKernel
        ) -> bool:
    """
    Returns *True* iff *sar1* and *sar2* are equivalent
    :class:`loopy.SubArrayRef`s.

    Two sub-array-refs are said to be equivalent iff they point to the same
    array sub-regions. This equivalence check is less strict than
    :meth:`~loopy.SubArrayRef.is_equal`.

    :arg caller: An instance of :class:`loopy.LoopKernel` in which they are
         referenced.
    """
    if len(sar1.swept_inames) != len(sar2.swept_inames):
        return False

    assert isinstance(sar1.subscript.aggregate, Variable)
    assert isinstance(sar2.subscript.aggregate, Variable)
    if sar1.subscript.aggregate.name != sar2.subscript.aggregate.name:
        return False

    if len(sar1.subscript.index_tuple) != len(sar2.subscript.index_tuple):
        return False

    if (_get_sub_array_ref_swept_range(caller, sar1)
            != _get_sub_array_ref_swept_range(caller, sar2)):
        return False

    from pymbolic.mapper.substitutor import make_subst_func

    from loopy.symbolic import SubstitutionMapper, simplify_via_aff
    subst_func = make_subst_func({iname1.name:  iname2
                                  for iname1, iname2 in zip(sar1.swept_inames,
                                                            sar2.swept_inames)
                                  })

    # subst_mapper: maps swept inames from sar1 to sar2
    subst_mapper = SubstitutionMapper(subst_func)

    for idx1, idx2 in zip(sar1.subscript.index_tuple,
                          sar2.subscript.index_tuple):
        subst_idx1 = subst_mapper(idx1)
        assert is_arithmetic_expression(subst_idx1)
        assert is_arithmetic_expression(idx2)
        if simplify_via_aff(subst_idx1 - idx2) != 0:
            return False
    return True


def _validate_kernel_call_insn(
            caller: LoopKernel,
            call_insn: CallInstruction,
            callee: LoopKernel
        ) -> None:
    assert call_insn.expression.function.name == callee.name
    from loopy.kernel.array import ArrayBase
    from loopy.symbolic import SubArrayRef

    arg_id_to_arg = call_insn.arg_id_to_arg()

    next_iarg_input = 0
    next_iarg_output = -1

    for arg in callee.args:
        if arg.is_input:
            if next_iarg_input not in arg_id_to_arg:
                raise LoopyError(f"Call to '{callee.name}' in '{call_insn}' expects"
                                 f" a {next_iarg_input+1}-th positional "
                                 "argument corresponding"
                                 f" to '{arg.name}' in the callee.")
            in_val = arg_id_to_arg[next_iarg_input]
            next_iarg_input += 1
            if isinstance(arg, ArrayBase):
                if not isinstance(in_val, SubArrayRef):
                    raise LoopyError(f"Call to '{callee.name}' in '{call_insn}'"
                                     f" expects a sub-array-ref for '{arg.name}'"
                                     f" (got {in_val}).")
            else:
                if isinstance(in_val, SubArrayRef):
                    raise LoopyError(f"Call to '{callee.name}' in '{call_insn}'"
                                     f" expects a value argument for '{arg.name}'"
                                     f" (got {in_val}).")
        if arg.is_output:
            if next_iarg_output not in arg_id_to_arg:
                raise LoopyError(f"Call to '{callee.name}' in '{call_insn}' expects"
                                 f" a {-next_iarg_output}-th positional assignee"
                                 f" corresponding to '{arg.name}'in the callee.")

            out_val = arg_id_to_arg[next_iarg_output]
            next_iarg_output -= 1
            assert isinstance(arg, ArrayBase)
            if not isinstance(out_val, SubArrayRef):
                raise LoopyError(f"Call to '{callee.name}' in '{call_insn}'"
                                 f" expects a sub-array-ref for '{arg.name}'"
                                 f" (got {out_val}).")

        if arg.is_input and arg.is_output:
            if not _are_sub_array_refs_equivalent(in_val, out_val, caller):
                raise LoopyError(f"Call to '{callee.name}' in '{call_insn}' expects"
                                 f" equivalent sub-array-refs for '{arg.name}'"
                                 f" (got {in_val}, {out_val}).")


def _validate_kernel_call_sites_inner(
            kernel: LoopKernel,
            callables: CallablesTable,
        ) -> None:
    from pymbolic.primitives import Call

    from loopy.kernel.function_interface import CallableKernel

    for insn in kernel.instructions:
        if (isinstance(insn, CallInstruction)
                and isinstance(insn.expression, Call)
                and isinstance(insn.expression.function, ResolvedFunction)):
            clbl = callables[insn.expression.function.name]
            if isinstance(clbl, CallableKernel):
                _validate_kernel_call_insn(kernel, insn, clbl.subkernel)
        elif isinstance(insn, (MultiAssignmentBase, CInstruction,
                               _DataObliviousInstruction)):
            pass
        else:
            raise NotImplementedError(type(insn))


def validate_kernel_call_sites(translation_unit: TranslationUnit) -> None:
    for name in translation_unit.callables_table:
        clbl = translation_unit.callables_table[name]
        if isinstance(clbl, CallableKernel):
            _validate_kernel_call_sites_inner(
                  clbl.subkernel, translation_unit.callables_table)


# }}}


# {{{ check_all_callees_have_same_index_dtype

def check_all_callees_have_same_index_dtype(
            epoint: LoopKernel,
            callables_table: CallablesTable
        ) -> None:
    from loopy.kernel.function_interface import CallableKernel

    epoint_clbl = callables_table[epoint.name]
    for clbl_name in epoint_clbl.get_called_callables(callables_table,
                                                      recursive=True):
        clbl = callables_table[clbl_name]
        if isinstance(clbl, CallableKernel):
            if clbl.subkernel.index_dtype != epoint.index_dtype:
                raise LoopyError(f"Callee '{clbl_name}' in entrypoint"
                                 f" {epoint.name}'s callgraph does not have the"
                                 " same index_dtype as the entrypoint. This is"
                                 " not allowed.")

# }}}


def pre_codegen_entrypoint_checks(
            kernel: LoopKernel,
            callables_table: CallablesTable
        ) -> None:
    logger.debug("pre-codegen entrypoint check %s: start" % kernel.name)

    kernel.target.pre_codegen_entrypoint_check(kernel, callables_table)
    check_all_callees_have_same_index_dtype(kernel, callables_table)

    logger.debug("pre-codegen entrypoint check %s: done" % kernel.name)


def pre_codegen_callable_checks(kernel, callables_table):
    logger.debug("pre-codegen callable check %s: start" % kernel.name)

    check_for_nested_base_storage(kernel)
    check_for_unused_hw_axes_in_insns(kernel, callables_table)
    check_that_atomic_ops_are_used_exactly_on_atomic_arrays(kernel)
    check_that_temporaries_are_defined_in_subkernels_where_used(kernel)
    check_that_all_insns_are_scheduled(kernel)
    kernel.target.pre_codegen_callable_check(kernel, callables_table)
    check_that_shapes_and_strides_are_arguments(kernel)

    logger.debug("pre-codegen callable check %s: done" % kernel.name)


def pre_codegen_checks(t_unit: TranslationUnit) -> None:
    from loopy.kernel.function_interface import CallableKernel

    try:
        for e in t_unit.entrypoints:
            pre_codegen_entrypoint_checks(t_unit[e], t_unit.callables_table)

        for clbl in t_unit.callables_table.values():
            if isinstance(clbl, CallableKernel):
                pre_codegen_callable_checks(clbl.subkernel, t_unit.callables_table)
    except Exception:
        print(75*"=")
        print("failing kernel during pre-codegen check:")
        print(75*"=")
        print(t_unit)
        print(75*"=")
        raise

# }}}


# {{{ sanity-check for implemented domains of each instruction

def check_implemented_domains(
            kernel: LoopKernel,
            implemented_domains: Mapping[str, isl.Set],
            code: str | None = None,
        ) -> bool:
    from islpy import align_two, dim_type

    last_idomains = None
    last_insn_inames = None

    for insn_id, idomains in implemented_domains.items():
        insn = kernel.id_to_insn[insn_id]

        assert idomains

        insn_inames = insn.within_inames

        # {{{ if we've checked the same thing before, no need to check it again

        if last_idomains is not None and last_insn_inames is not None:
            if idomains == last_idomains and insn_inames == last_insn_inames:
                continue

        last_idomains = idomains
        last_insn_inames = insn_inames

        # }}}

        insn_impl_domain = idomains[0]
        for idomain in idomains[1:]:
            insn_impl_domain = insn_impl_domain | idomain
        assumption_non_param = isl.BasicSet.from_params(kernel.assumptions)
        assumptions, insn_impl_domain = align_two(
                assumption_non_param, insn_impl_domain)
        insn_impl_domain = (
                (insn_impl_domain & assumptions)
                .project_out_except(insn_inames, [dim_type.set]))

        from loopy.kernel.data import LocalInameTag
        from loopy.kernel.instruction import BarrierInstruction
        if isinstance(insn, BarrierInstruction):
            # project out local-id-mapped inames, solves #94 on gitlab
            non_lid_inames = frozenset(iname for iname in insn_inames
                if not kernel.iname_tags_of_type(iname, LocalInameTag))
            insn_impl_domain = insn_impl_domain.project_out_except(
                non_lid_inames, [dim_type.set])

        insn_domain = kernel.get_inames_domain(insn_inames)
        insn_parameters = frozenset(insn_domain.get_var_names(dim_type.param))
        assumptions, insn_domain = align_two(assumption_non_param, insn_domain)
        desired_domain = ((insn_domain & assumptions)
            .project_out_except(insn_inames, [dim_type.set])
            .project_out_except(insn_parameters, [dim_type.param]))

        if isinstance(insn, BarrierInstruction):
            # project out local-id-mapped inames, solves #94 on gitlab
            desired_domain = desired_domain.project_out_except(
                non_lid_inames, [dim_type.set])

        insn_impl_domain = (insn_impl_domain
                .project_out_except(insn_parameters, [dim_type.param]))
        insn_impl_domain, desired_domain = align_two(
                insn_impl_domain, desired_domain)

        if insn_impl_domain != desired_domain:
            i_minus_d = insn_impl_domain - desired_domain
            d_minus_i = desired_domain - insn_impl_domain

            parameter_inames = {
                    insn_domain.get_dim_name(dim_type.param, i)
                    for i in range(insn_impl_domain.dim(dim_type.param))}

            lines = []
            for bigger, smaller, diff_set, gist_domain in [
                    ("implemented", "desired", i_minus_d,
                        desired_domain.gist(insn_impl_domain)),
                    ("desired", "implemented", d_minus_i,
                        insn_impl_domain.gist(desired_domain))]:

                if diff_set.is_empty():
                    continue

                diff_set = diff_set.coalesce()
                pt = diff_set.sample_point()
                assert not pt.is_void()

                # pt_set = isl.Set.from_point(pt)
                # lines.append("point implemented: %s" % (pt_set <= insn_impl_domain))
                # lines.append("point desired: %s" % (pt_set <= desired_domain))

                iname_to_dim = pt.get_space().get_var_dict()
                point_axes = []
                for iname in insn_inames | parameter_inames:
                    tp, dim = iname_to_dim[iname]
                    point_axes.append("%s=%d" % (
                        iname, pt.get_coordinate_val(tp, dim).to_python()))

                lines.append(
                        "sample point in {} but not {}: {}".format(
                            bigger, smaller, ", ".join(point_axes)))
                lines.append(
                        "gist of constraints in {} but not {}: {}".format(
                            bigger, smaller, gist_domain))

            if code is not None:
                print(79*"-")
                print("CODE:")
                print(79*"-")
                from loopy.target.execution import get_highlighted_code
                print(get_highlighted_code(code))
                print(79*"-")

            raise LoopyError("sanity check failed--implemented and desired "
                    "domain for instruction '%s' do not match\n\n"
                    "implemented: %s\n\n"
                    "desired:%s\n\n%s"
                    % (insn_id, insn_impl_domain, desired_domain, "\n".join(lines)))

    # placate the assert at the call site
    return True

# }}}


# vim: foldmethod=marker



from __future__ import annotations


__copyright__ = "Copyright (C) 2012 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from pytools import MovedFunctionDeprecationWrapper


# {{{ warnings

class LoopyWarningBase(UserWarning):
    pass


class LoopyWarning(LoopyWarningBase):
    pass


class LoopyAdvisory(LoopyWarningBase):
    pass


class ParameterFinderWarning(LoopyWarning):
    pass


class WriteRaceConditionWarning(LoopyWarning):
    pass


class DirectCallUncachedWarning(LoopyWarning):
    pass

# }}}


def warn_with_kernel(kernel, id, text, type=LoopyWarning, stacklevel=None):
    from fnmatch import fnmatchcase
    for sw in kernel.silenced_warnings:
        if fnmatchcase(id, sw):
            return

    text += (" (add '%s' to silenced_warnings kernel attribute to disable)"
            % id)

    if stacklevel is None:
        stacklevel = 2
    else:
        stacklevel = stacklevel + 1
    from warnings import warn
    warn(f"in kernel {kernel.name}: {text}", type, stacklevel=stacklevel)


warn = MovedFunctionDeprecationWrapper(warn_with_kernel)


# {{{ errors

class LoopyError(RuntimeError):
    pass


class LoopyIndexError(LoopyError):
    pass


class CannotBranchDomainTree(LoopyError):  # noqa: N818
    pass


class TypeInferenceFailure(LoopyError):  # noqa: N818
    pass


class AutomaticTestFailure(LoopyError):  # noqa: N818
    pass


class StaticValueFindingError(LoopyError):
    pass


class DependencyTypeInferenceFailure(TypeInferenceFailure):
    pass


class MissingBarrierError(LoopyError):
    pass


class MissingDefinitionError(LoopyError):
    pass


class UnscheduledInstructionError(LoopyError):
    pass


class ReductionIsNotTriangularError(LoopyError):
    pass


class LoopyTypeError(LoopyError):
    pass


class ExpressionNotAffineError(LoopyError):
    """
    Raised when an expression is not quasi-affine. See
    `ISL manual <http://isl.gforge.inria.fr//user.html#Primitive-Functions>`_
    for then definition of a quasi-affine expression.
    """
    pass


class ExpressionToAffineConversionError(LoopyError):
    pass


class VariableAccessNotOrdered(LoopyError):  # noqa: N818
    pass


class DependencyCycleFound(LoopyError):  # noqa: N818
    pass


class UnableToDetermineAccessRangeError(Exception):
    pass


class ScheduleDebugInputError(Exception):
    pass

# }}}


# vim: foldmethod=marker



"""
.. autoclass:: Expression
.. autoclass:: ShapeType
.. autodata:: InameStr
.. autodata:: InameStrSet

.. currentmodule:: loopy

.. autoclass:: auto
"""


from __future__ import annotations


__copyright__ = "Copyright (C) 2022 University of Illinois Board of Trustees"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


from typing import Tuple, TypeVar

import numpy as np
from typing_extensions import TypeAlias, TypeIs

from pymbolic.primitives import ExpressionNode
from pymbolic.typing import ArithmeticExpression, Expression, Integer


# The Fortran parser may insert dimensions of 'None', but I'd like to phase
# that out, so we're not encoding that in the type.
ShapeType: TypeAlias = Tuple[ArithmeticExpression, ...]
StridesType: TypeAlias = ShapeType

InameStr: TypeAlias = str
InameStrSet: TypeAlias = frozenset[InameStr]


class auto:  # noqa
    """A generic placeholder object for something that should be automatically
    determined.  See, for example, the *shape* or *strides* argument of
    :class:`~loopy.ArrayArg`.
    """


T = TypeVar("T")


def not_none(obj: T | None) -> T:
    assert obj is not None
    return obj


def is_integer(obj: object) -> TypeIs[int | np.integer]:
    return isinstance(obj, (int, np.integer))


def integer_or_err(expr: Expression) -> Integer:
    if isinstance(expr, (int, np.integer)):
        return expr
    else:
        raise ValueError(f"expected integer, got {type(expr)}")


def integer_expr_or_err(expr: Expression) -> Integer | ExpressionNode:
    if isinstance(expr, (int, np.integer, ExpressionNode)):
        return expr
    else:
        raise ValueError(f"expected integer or expression, got {type(expr)}")


ElT = TypeVar("ElT")


def assert_tuple(obj: tuple[ElT, ...] | object) -> tuple[ElT, ...]:
    assert isinstance(obj, tuple)
    return obj



from __future__ import annotations


__copyright__ = "Copyright (C) 2013 Andreas Kloeckner"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


import os
import re
from typing import TYPE_CHECKING, Any, ClassVar
from warnings import warn

from pytools import ImmutableRecord


if TYPE_CHECKING:
    from collections.abc import Mapping


ALLOW_TERMINAL_COLORS = True


class _ColoramaStub:
    def __getattribute__(self, name):
        return ""


def _apply_legacy_map(lmap, kwargs):
    result = {}

    for name, val in kwargs.items():
        try:
            lmap_value = lmap[name]
        except KeyError:
            new_name = name
        else:
            if lmap_value is None:
                # ignore this
                warn("option '%s' is deprecated and was ignored" % name,
                        DeprecationWarning, stacklevel=1)
                continue

            new_name, translator = lmap_value
            if name in result:
                raise TypeError("may not pass a value for both '%s' and '%s'"
                        % (name, new_name))

            warn(f"Loopy option '{name}' is deprecated. '{new_name}' should be "
                    "used instead. The old option will stop working in 2022.",
                    DeprecationWarning, stacklevel=1)
            if translator is not None:
                val = translator(val)

        result[new_name] = val

    return result


class Options(ImmutableRecord):
    """
    Unless otherwise specified, these options are Boolean-valued
    (i.e. on/off).

    .. rubric:: Code-generation options

    .. attribute:: annotate_inames

        When generating code for inames, annotate them with
        comments if it is not immediately apparent which
        iname is being referred to (such as for inames mapped
        to constants or OpenCL group/local IDs).

    .. attribute:: trace_assignments

        Generate code that uses *printf* in kernels to trace the
        execution of assignment instructions.

    .. attribute:: trace_assignment_values

        Like :attr:`trace_assignments`, but also trace the
        assigned values.

    .. attribute:: check_dep_resolution

        Whether loopy should issue an error if a dependency
        expression does not match any instructions in the kernel.

    .. rubric:: Invocation-related options

    .. attribute:: skip_arg_checks

        Do not do any checking (data type, data layout, shape,
        etc.) on arguments for a minor performance gain.

        .. versionchanged:: 2021.1

            This now defaults to the same value as the ``optimize``
            sub-flag from :data:`sys.flags`. This flag can be controlled
            (i.e. set to *True*) by running Python with the ``-O`` flag.

    .. attribute:: no_numpy

        Do not check for or accept :mod:`numpy` arrays as
        arguments.

        Defaults to *False*.

    .. attribute:: cl_exec_manage_array_events

        Within the PyOpenCL executor, respect and update
        :attr:`pyopencl.array.Array.events`.

        Defaults to *True*.

    .. attribute:: return_dict

        Have kernels return a :class:`dict` instead of a tuple as
        output. Specifically, the result of a kernel invocation
        with this flag is a tuple ``(evt, out_dict)``, where
        *out_dict* is a dictionary mapping argument names to their
        output values. This is helpful if arguments are inferred
        and argument ordering is thus implementation-defined.

        See :meth:`ExecutorBase.__call__`.

    .. attribute:: write_wrapper

        Print the generated Python invocation wrapper.
        Accepts a file name as a value. Writes to
        ``sys.stdout`` if none is given.

    .. attribute:: write_code

        Print the generated code.  Accepts a file name or a boolean as a value.
        Writes to ``sys.stdout`` if set to *True*.

    .. attribute:: edit_code

        Invoke an editor (given by the environment variable
        ``EDITOR``) on the generated kernel code,
        allowing for tweaks before the code is passed on to
        the target for compilation.

    .. attribute:: allow_fp_reordering

        Allow re-ordering of floating point arithmetic. Re-ordering may
        give different results as floating point arithmetic is not
        associative in addition and multiplication. Default is *True*.
        Note that the implementation of this option is currently incomplete.

    .. attribute:: build_options

        Options to pass to the target compiler when building the kernel.
        A list of strings.

    .. attribute:: allow_terminal_colors

        A :class:`bool`. Whether to allow colors in terminal output

    .. rubric:: Features

    .. attribute:: disable_global_barriers

    .. attribute:: enforce_variable_access_ordered

        If *True*, require that
        :func:`loopy.check.check_variable_access_ordered` passes.
        Required for language versions 2018.1 and above. This check
        helps find and eliminate unintentionally unordered access
        to variables.

        If equal to ``"no_check"``, then no check is performed.

    .. attribute:: enforce_array_accesses_within_bounds

        If *True*, require that :func:`~loopy.check.check_bounds` passes. If
        *False*, then :func:`~loopy.check.check_bounds` raises a warning for
        any out-of-bounds accesses.

        If equal to ``"no_check"``, then no check is performed.

    .. attribute:: insert_gbarriers

        If *True*, based on the memory dependency between variables in the
        global address space loopy will insert global barriers to avoid
        RAW, WAR and WAW races.
    """

    _legacy_options_map: ClassVar[Mapping[str, tuple[str, None] | None]] = {
            "cl_build_options": ("build_options", None),
            "write_cl": ("write_code", None),
            "highlight_cl": None,
            "highlight_wrapper": None,
            "disable_wrapper_highlight": None,
            "disable_code_highlight": None,
            "edit_cl": ("edit_code", None),
            }

    def __init__(
            # All Boolean flags in here should default to False for the
            # string-based interface of make_options (below) to make sense.

            # All defaults are further required to be False when cast to bool
            # for the update() functionality to work.

            self, **kwargs: Any) -> None:

        kwargs = _apply_legacy_map(self._legacy_options_map, kwargs)

        try:
            import colorama  # noqa: F401
        except ImportError:
            allow_terminal_colors_def = False
        else:
            allow_terminal_colors_def = True

        allow_terminal_colors_def = (
                ALLOW_TERMINAL_COLORS
                and allow_terminal_colors_def
                # https://no-color.org/
                and "NO_COLOR" not in os.environ)

        import sys
        ImmutableRecord.__init__(
                self,

                annotate_inames=kwargs.get("annotate_inames", False),
                trace_assignments=kwargs.get("trace_assignments", False),
                trace_assignment_values=kwargs.get("trace_assignment_values", False),

                skip_arg_checks=kwargs.get("skip_arg_checks",
                    sys.flags.optimize
                    # Not considered a documented env var: Only used to test
                    # the skip_arg_checks branch during CI, which can't use
                    # python -O.
                    #
                    # Considered enabled if non-empty.
                    or bool(os.environ.get("_LOOPY_SKIP_ARG_CHECKS"))),
                no_numpy=kwargs.get("no_numpy", False),
                cl_exec_manage_array_events=kwargs.get("cl_exec_manage_array_events",
                    True),
                return_dict=kwargs.get("return_dict", False),
                write_wrapper=kwargs.get("write_wrapper", False),
                write_code=kwargs.get("write_code", False),
                edit_code=kwargs.get("edit_code", False),
                allow_fp_reordering=kwargs.get("allow_fp_reordering", True),
                build_options=kwargs.get("build_options", []),
                allow_terminal_colors=kwargs.get("allow_terminal_colors",
                    allow_terminal_colors_def),
                disable_global_barriers=kwargs.get("disable_global_barriers",
                    False),
                check_dep_resolution=kwargs.get("check_dep_resolution", True),

                enforce_variable_access_ordered=kwargs.get(
                    "enforce_variable_access_ordered", True),
                enforce_array_accesses_within_bounds=kwargs.get(
                    "enforce_array_accesses_within_bounds", True),
                insert_gbarriers=kwargs.get(
                    "insert_gbarriers", False),
                )

    # {{{ legacy compatibility

    @property
    def edit_cl(self):
        return self.edit_code

    @property
    def cl_build_options(self):
        return self.build_options

    @property
    def highlight_cl(self):
        return self.allow_terminal_colors

    @property
    def highlight_wrapper(self):
        return self.allow_terminal_colors

    @property
    def write_cl(self):
        return self.write_code

    # }}}

    # only used internally on new copies of Options
    def _update(self, other):
        for f in self.__class__.fields:
            setattr(self, f, getattr(self, f) or getattr(other, f))

    def update_persistent_hash(self, key_hash, key_builder):
        """Custom hash computation function for use with
        :class:`pytools.persistent_dict.PersistentDict`.
        """
        for field_name in sorted(self.__class__.fields):
            key_builder.rec(key_hash, getattr(self, field_name))

    @property
    def _fore(self):
        if self.allow_terminal_colors:
            import colorama
            return colorama.Fore
        else:
            return _ColoramaStub()

    @property
    def _back(self):
        if self.allow_terminal_colors:
            import colorama
            return colorama.Back
        else:
            return _ColoramaStub()

    @property
    def _style(self):
        if self.allow_terminal_colors:
            import colorama
            return colorama.Style
        else:
            return _ColoramaStub()


KEY_VAL_RE = re.compile(r"^([a-zA-Z0-9]+)=(.*)$")


def make_options(options_arg):
    if options_arg is None:
        return Options()
    elif isinstance(options_arg, str):
        ioptions_args = {}
        for key_val in options_arg.split(","):
            kv_match = KEY_VAL_RE.match(key_val)
            if kv_match is not None:
                key = kv_match.group(1)
                val = kv_match.group(2)
                try:
                    val = int(val)
                except ValueError:
                    pass

                ioptions_args[key] = val
            else:
                ioptions_args[key_val] = True

        return Options(**ioptions_args)
    elif not isinstance(options_arg, Options):
        return Options(**options_arg)
    elif isinstance(options_arg, Options):
        return options_arg
    else:
        raise TypeError("invalid argument to make_options")



from __future__ import annotations


__copyright__ = "Copyright (C) 2018 Tianjiao Sun, Kaushik Kulkarni"

__license__ = """
Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from constantdict import constantdict

from loopy.diagnostic import LoopyError
from loopy.kernel import LoopKernel
from loopy.kernel.function_interface import CallableKernel, ScalarCallable
from loopy.kernel.instruction import CallInstruction
from loopy.symbolic import SubArrayRef
from loopy.translation_unit import TranslationUnit


__doc__ = """
.. currentmodule:: loopy

.. autofunction:: pack_and_unpack_args_for_call
"""


def pack_and_unpack_args_for_call_for_single_kernel(kernel,
        callables_table, call_name, args_to_pack=None,
        args_to_unpack=None):
    """
    Returns a a copy of *kernel* with instructions appended to copy the
    arguments in *args* to match the alignment expected by the *call_name* in
    the kernel. The arguments are copied back to *args* with the appropriate
    data layout.

    :arg call_name: An instance of :class:`str` denoting the function call in
        the *kernel*.
    :arg args_to_pack: A list of the arguments as instances of :class:`str` which
        must be packed. If set *None*, it is interpreted that all the array
        arguments would be packed.
    :arg args_to_unpack: A list of the arguments as instances of :class:`str`
        which must be unpacked. If set *None*, it is interpreted that
        all the array arguments should be unpacked.
    """
    assert isinstance(kernel, LoopKernel)
    new_domains = []
    new_tmps = kernel.temporary_variables.copy()
    old_insn_to_new_insns = {}

    for insn in kernel.instructions:
        if not isinstance(insn, CallInstruction):
            # pack and unpack call only be done for CallInstructions.
            continue
        if insn.expression.function.name not in callables_table:
            continue

        in_knl_callable = callables_table[
                insn.expression.function.name]

        if in_knl_callable.name != call_name:
            # not the function we're looking for.
            continue
        in_knl_callable = in_knl_callable.with_packing_for_args()

        vng = kernel.get_var_name_generator()
        ing = kernel.get_instruction_id_generator()

        parameters = insn.expression.parameters
        if args_to_pack is None:
            args_to_pack = [par.subscript.aggregate.name for par in
                    parameters+insn.assignees if isinstance(par, SubArrayRef)
                    and (par.swept_inames)]
        if args_to_unpack is None:
            args_to_unpack = [par.subscript.aggregate.name for par in
                    parameters+insn.assignees if isinstance(par, SubArrayRef)
                    and (par.swept_inames)]

        # {{{ sanity checks for args

        assert isinstance(args_to_pack, list)
        assert isinstance(args_to_unpack, list)

        for arg in args_to_pack:
            found_sub_array_ref = False

            for par in parameters + insn.assignees:
                # checking that the given args is a sub array ref
                if isinstance(par, SubArrayRef) and (
                        par.subscript.aggregate.name == arg):
                    found_sub_array_ref = True
                    break
            if not found_sub_array_ref:
                raise LoopyError("No match found for packing arg '%s' of call '%s' "
                        "at insn '%s'." % (arg, call_name, insn.id))
        for arg in args_to_unpack:
            if arg not in args_to_pack:
                raise LoopyError("Argument %s should be packed in order to be "
                        "unpacked." % arg)

        # }}}

        packing_insns = []
        unpacking_insns = []

        # {{{ handling ilp tags

        import islpy as isl
        from pymbolic import var

        from loopy.kernel.data import IlpBaseTag, VectorizeTag

        dim_type = isl.dim_type.set
        ilp_inames = {iname for iname in insn.within_inames
                         if all(isinstance(tag, (IlpBaseTag, VectorizeTag))
                                for tag in kernel.iname_to_tags.get(iname, []))}
        new_ilp_inames = set()
        ilp_inames_map = {}
        for iname in ilp_inames:
            new_iname_name = vng(iname + "_ilp")
            ilp_inames_map[var(iname)] = var(new_iname_name)
            new_ilp_inames.add(new_iname_name)
        for iname in ilp_inames:
            new_domain = kernel.get_inames_domain(iname).copy()
            for i in range(new_domain.n_dim()):
                old_iname = new_domain.get_dim_name(dim_type, i)
                if old_iname in ilp_inames:
                    new_domain = new_domain.set_dim_name(
                        dim_type, i, ilp_inames_map[var(old_iname)].name)
            new_domains.append(new_domain)

        # }}}

        from pymbolic.mapper.substitutor import make_subst_func

        from loopy.symbolic import SubstitutionMapper

        # dict to store the new assignees and parameters, the mapping pattern
        # from arg_id to parameters is identical to InKernelCallable.arg_id_to_dtype
        id_to_parameters = tuple(enumerate(parameters)) + tuple(
                (-i-1, assignee) for i, assignee in enumerate(insn.assignees))
        new_id_to_parameters = {}

        for arg_id, p in id_to_parameters:
            if isinstance(p, SubArrayRef) and (p.subscript.aggregate.name in
                    args_to_pack):
                new_pack_inames = ilp_inames_map.copy()  # packing-specific inames
                new_unpack_inames = ilp_inames_map.copy()  # unpacking-specific iname

                new_pack_inames = {iname: var(vng(iname.name +
                    "_pack")) for iname in p.swept_inames}
                new_unpack_inames = {iname: var(vng(iname.name +
                    "_unpack")) for iname in p.swept_inames}

                # Updating the domains corresponding to the new inames.
                for iname in p.swept_inames:
                    new_domain_pack = kernel.get_inames_domain(iname.name).copy()
                    new_domain_unpack = kernel.get_inames_domain(iname.name).copy()
                    for i in range(new_domain_pack.n_dim()):
                        old_iname = new_domain_pack.get_dim_name(dim_type, i)
                        if var(old_iname) in new_pack_inames:
                            new_domain_pack = new_domain_pack.set_dim_name(
                                dim_type, i, new_pack_inames[var(old_iname)].name)
                            new_domain_unpack = new_domain_unpack.set_dim_name(
                                dim_type, i, new_unpack_inames[var(old_iname)].name)
                    new_domains.append(new_domain_pack)
                    new_domains.append(new_domain_unpack)

                arg = p.subscript.aggregate.name
                pack_name = vng(arg + "_pack")

                from loopy.kernel.data import AddressSpace, TemporaryVariable

                if arg in kernel.arg_dict:
                    arg_in_caller = kernel.arg_dict[arg]
                else:
                    arg_in_caller = kernel.temporary_variables[arg]

                pack_tmp = TemporaryVariable(
                    name=pack_name,
                    dtype=arg_in_caller.dtype,
                    dim_tags=in_knl_callable.arg_id_to_descr[arg_id].dim_tags,
                    shape=in_knl_callable.arg_id_to_descr[arg_id].shape,
                    address_space=AddressSpace.PRIVATE,
                )

                new_tmps[pack_name] = pack_tmp

                from loopy import Assignment
                pack_subst_mapper = SubstitutionMapper(make_subst_func(
                    new_pack_inames))
                unpack_subst_mapper = SubstitutionMapper(make_subst_func(
                    new_unpack_inames))

                # {{{ getting the lhs for packing and rhs for unpacking

                from loopy.isl_helpers import make_slab
                from loopy.symbolic import simplify_via_aff

                flatten_index = simplify_via_aff(
                        sum(dim_tag.stride*idx for dim_tag, idx in
                        zip(arg_in_caller.dim_tags, p.subscript.index_tuple)))

                new_indices = []
                for dim_tag in in_knl_callable.arg_id_to_descr[arg_id].dim_tags:
                    ind = flatten_index // dim_tag.stride
                    flatten_index -= (dim_tag.stride * ind)
                    new_indices.append(ind)

                new_indices = tuple(simplify_via_aff(i) for i in new_indices)

                pack_lhs_assignee = pack_subst_mapper(
                        var(pack_name)[new_indices])
                unpack_rhs = unpack_subst_mapper(
                        var(pack_name)[new_indices])

                # }}}

                packing_insns.append(Assignment(
                    assignee=pack_lhs_assignee,
                    expression=pack_subst_mapper.map_subscript(p.subscript),
                    within_inames=insn.within_inames - ilp_inames | {
                        new_pack_inames[i].name for i in p.swept_inames} | (
                            new_ilp_inames),
                    depends_on=insn.depends_on,
                    id=ing(insn.id+"_pack"),
                    depends_on_is_final=True
                ))

                if p.subscript.aggregate.name in args_to_unpack:
                    unpacking_insns.append(Assignment(
                        expression=unpack_rhs,
                        assignee=unpack_subst_mapper.map_subscript(p.subscript),
                        within_inames=insn.within_inames - ilp_inames | {
                            new_unpack_inames[i].name for i in p.swept_inames} | (
                                new_ilp_inames),
                        id=ing(insn.id+"_unpack"),
                        depends_on=frozenset([insn.id]),
                        depends_on_is_final=True
                    ))

                # {{{ creating the sweep inames for the new sub array refs

                updated_swept_inames = []

                for _ in in_knl_callable.arg_id_to_descr[arg_id].shape:
                    updated_swept_inames.append(var(vng("i_packsweep_"+arg)))

                ctx = kernel.isl_context
                space = isl.Space.create_from_names(ctx,
                        set=[iname.name for iname in updated_swept_inames])
                iname_set = isl.BasicSet.universe(space)
                for iname, axis_length in zip(updated_swept_inames,
                        in_knl_callable.arg_id_to_descr[arg_id].shape):
                    iname_set = iname_set & make_slab(space, iname.name, 0,
                            axis_length)
                new_domains = [*new_domains, iname_set]

                # }}}

                new_id_to_parameters[arg_id] = SubArrayRef(
                        tuple(updated_swept_inames),
                        (var(pack_name)[tuple(updated_swept_inames)]))
            else:
                new_id_to_parameters[arg_id] = p

        if packing_insns:
            subst_mapper = SubstitutionMapper(make_subst_func(ilp_inames_map))
            new_call_insn = insn.with_transformed_expressions(subst_mapper)
            new_params = tuple(subst_mapper(new_id_to_parameters[i]) for i, _ in
                    enumerate(parameters))
            new_assignees = tuple(subst_mapper(new_id_to_parameters[-i-1])
                    for i, _ in enumerate(insn.assignees))
            new_call_insn = new_call_insn.copy(
                    depends_on=new_call_insn.depends_on | {
                        pack.id for pack in packing_insns},
                    within_inames=new_call_insn.within_inames - ilp_inames | (
                        new_ilp_inames),
                    expression=new_call_insn.expression.function(*new_params),
                    assignees=new_assignees)
            old_insn_to_new_insns[insn.id] = ([
                *packing_insns, new_call_insn, *unpacking_insns])

    if old_insn_to_new_insns:
        new_instructions = []
        for insn in kernel.instructions:
            if insn.id in old_insn_to_new_insns:
                # Replacing the current instruction with the group of
                # instructions including the packing and unpacking instructions
                new_instructions.extend(old_insn_to_new_insns[insn.id])
            else:
                # for the instructions that depend on the call instruction that
                # are to be packed and unpacked, we need to add the complete
                # instruction block as a dependency for them.
                new_depends_on = insn.depends_on
                if insn.depends_on & set(old_insn_to_new_insns):
                    # need to add the unpack instructions on dependencies.
                    for old_insn_id in insn.depends_on & set(old_insn_to_new_insns):
                        new_depends_on |= frozenset(i.id for i
                                in old_insn_to_new_insns[old_insn_id])
                new_instructions.append(insn.copy(depends_on=new_depends_on))
        kernel = kernel.copy(
            domains=kernel.domains + new_domains,
            instructions=new_instructions,
            temporary_variables=new_tmps
        )

    return kernel


def pack_and_unpack_args_for_call(program, *args, **kwargs):
    assert isinstance(program, TranslationUnit)

    new_callables = {}
    for func_id, in_knl_callable in program.callables_table.items():
        if isinstance(in_knl_callable, CallableKernel):
            new_subkernel = pack_and_unpack_args_for_call_for_single_kernel(
                    in_knl_callable.subkernel, program.callables_table,
                    *args, **kwargs)
            in_knl_callable = in_knl_callable.copy(
                    subkernel=new_subkernel)
        elif isinstance(in_knl_callable, ScalarCallable):
            pass
        else:
            raise NotImplementedError("Unknown type of callable %s." % (
                type(in_knl_callable).__name__))

        new_callables[func_id] = in_knl_callable

    return program.copy(callables_table=constantdict(new_callables))

# vim: foldmethod=marker
